import json as _json
from arcgis.raster._layer import ImageryLayer as _ImageryLayer

# from arcgis.raster._layer import Raster as _Raster
from arcgis.features import FeatureLayer as _FeatureLayer
import arcgis as _arcgis
import string as _string
import random as _random
from arcgis._impl.common._utils import _date_handler
import datetime
from arcgis.geometry import Geometry as _Geometry
import numbers
import time
import os
from urllib.parse import urljoin, quote, unquote, urlparse
import sys


import logging as _logging

_LOGGER = _logging.getLogger(__name__)

try:
    import numpy as _np
    import requests as _requests
except:
    pass


def _get_layer_info(input_layer):
    input_param = input_layer

    url = ""
    from arcgis.raster import Raster as _Raster
    from arcgis.gis import Item as _Item
    from arcgis.gis import Layer as _Layer

    if isinstance(input_layer, _Raster):
        if hasattr(input_layer, "_engine_obj"):
            input_layer = input_layer._engine_obj
    if isinstance(input_layer, _Item):
        if input_layer.type == "Image Collection":
            input_param = {"itemId": input_layer.itemid}
        else:
            if "layers" in input_layer:
                input_param = input_layer.layers[0]._lyr_dict
                try:
                    if isinstance(input_param, dict) and "url" in input_param.keys():
                        url = input_param["url"]
                        if "token" not in url:
                            from arcgis.raster.functions.utility import (
                                _generate_layer_token,
                            )

                            token = _generate_layer_token(input_layer, url)
                            if token is not None:
                                if input_layer.type == "Feature Service":
                                    input_param.update({"serviceToken": token})
                                else:
                                    url = input_param["url"] + "?token=" + token
                            input_param.update({"url": url})
                except:
                    pass
            else:
                raise TypeError("No layers in input layer Item")

    elif isinstance(input_layer, _Layer):
        input_param = input_layer._lyr_dict
        from arcgis.raster import ImageryLayer
        import json

        if isinstance(input_layer, _ImageryLayer) or isinstance(input_layer, _Raster):
            if "options" in input_layer._lyr_json:
                if isinstance(
                    input_layer._lyr_json["options"], str
                ):  # sometimes the rendering info is a string
                    # load json
                    layer_options = json.loads(input_layer._lyr_json["options"])
                else:
                    layer_options = input_layer._lyr_json["options"]

                if "imageServiceParameters" in layer_options:
                    # get renderingRule and mosaicRule
                    input_param.update(layer_options["imageServiceParameters"])

            try:
                if isinstance(input_param, dict) and "url" in input_param.keys():
                    url = input_param["url"]
                    if "token" not in url:
                        from arcgis.raster.functions.utility import (
                            _generate_layer_token,
                        )

                        token = _generate_layer_token(input_layer, url)
                        if token is not None:
                            url = input_param["url"] + "?token=" + token
                        input_param.update({"url": url})
                        if "serviceToken" in input_param.keys():
                            del input_param["serviceToken"]
            except:
                pass

        elif isinstance(input_layer, _FeatureLayer):
            input_param = input_layer._lyr_dict
            try:
                if isinstance(input_param, dict) and "url" in input_param.keys():
                    url = input_param["url"]
                    if "serviceToken" not in input_param:
                        from arcgis.raster.functions.utility import (
                            _generate_layer_token,
                        )

                        token = _generate_layer_token(input_layer, url)
                        if token is not None:
                            input_param.update({"serviceToken": token})
                        input_param.update({"url": url})
            except:
                pass

    elif isinstance(input_layer, dict):
        input_param = input_layer

    elif isinstance(input_layer, str):
        if "http:" in input_layer or "https:" in input_layer:
            input_param = {"url": input_layer}
        else:
            input_param = {"uri": input_layer}

    else:
        raise Exception("Invalid format for env parameter")

    if "ImageServer" in url or "MapServer" in url:
        if "serviceToken" in input_param:
            url = url + "?token=" + input_param["serviceToken"]
            input_param.update({"url": url})

    return input_param


def _set_context(params, function_context=None):
    out_sr = _arcgis.env.out_spatial_reference
    process_sr = _arcgis.env.process_spatial_reference
    out_extent = _arcgis.env.analysis_extent
    mask = _arcgis.env.mask
    snap_raster = _arcgis.env.snap_raster
    cell_size = _arcgis.env.cell_size
    parallel_processing_factor = _arcgis.env.parallel_processing_factor

    context = {}

    if out_sr is not None:
        context["outSR"] = {"wkid": int(out_sr)}

    if out_extent is not None:
        context["extent"] = out_extent

    if process_sr is not None:
        context["processSR"] = {"wkid": int(process_sr)}

    if mask is not None:
        context["mask"] = _get_layer_info(mask)

    if cell_size is not None:
        if isinstance(cell_size, _ImageryLayer):
            context["cellSize"] = {"url": cell_size._url}
        elif isinstance(cell_size, str):
            if "http:" in cell_size or "https:" in cell_size:
                context["cellSize"] = {"url": cell_size}
            else:
                context["cellSize"] = cell_size
        else:
            context["cellSize"] = cell_size

    if snap_raster is not None:
        context["snapRaster"] = _get_layer_info(snap_raster)

    if parallel_processing_factor is not None:
        context["parallelProcessingFactor"] = parallel_processing_factor

    if function_context is not None:
        if context is not None:
            context.update({k: function_context[k] for k in function_context.keys()})

        else:
            context = function_context

    if context:
        params["context"] = _json.dumps(context)


def _id_generator(size=6, chars=_string.ascii_uppercase + _string.digits):
    return "".join(_random.choice(chars) for _ in range(size))


def _set_time_param(time):
    time_val = time
    if time is not None:
        if type(time) is list:
            if isinstance(time[0], datetime.datetime):
                if time[0].tzname() is None or time[0].tzname() != "UTC":
                    time[0] = time[0].astimezone(datetime.timezone.utc)
            if isinstance(time[1], datetime.datetime):
                if time[1].tzname() is None or time[1].tzname() != "UTC":
                    time[1] = time[1].astimezone(datetime.timezone.utc)
            starttime = _date_handler(time[0])
            endtime = _date_handler(time[1])
            if starttime is None:
                starttime = "null"
            if endtime is None:
                endtime = "null"
            time_val = "%s,%s" % (starttime, endtime)
        else:
            time_val = _date_handler(time)

    return time_val


def _to_datetime(dt):
    import datetime

    try:
        if dt < 0:
            return datetime.datetime(1970, 1, 1) + datetime.timedelta(
                seconds=(dt / 1000)
            )
        else:
            return datetime.datetime.fromtimestamp(
                dt / 1000, tz=datetime.timezone.utc
            ).replace(tzinfo=None)
    except:
        return dt


def _datetime2ole(date):
    # date = datetime.strptime(date, '%d-%b-%Y')
    import datetime

    OLE_TIME_ZERO = datetime.datetime(1899, 12, 30)
    delta = date - OLE_TIME_ZERO
    return float(delta.days) + (float(delta.seconds) / 86400)


def _ole2datetime(oledt):
    import datetime

    OLE_TIME_ZERO = datetime.datetime(1899, 12, 30, 0, 0, 0)
    try:
        return OLE_TIME_ZERO + datetime.timedelta(days=float(oledt))
    except:
        return datetime.datetime.fromtimestamp(
            oledt / 1000, tz=datetime.timezone.utc
        ).replace(tzinfo=None)


def _iso_to_datetime(timestamp):
    format_string = "%Y-%m-%dT%H:%M:%S%z"
    try:
        colon = timestamp[-3]
        colonless_timestamp = timestamp
        if colon == ":":
            colonless_timestamp = timestamp[:-3] + timestamp[-2:]
        dt_ob = datetime.datetime.strptime(colonless_timestamp, format_string)
        return dt_ob.replace(tzinfo=None)
    except:
        try:
            format_string = "%Y-%m-%dT%H:%M:%S"
            dt_ob = datetime.datetime.strptime(timestamp, format_string)
            return dt_ob
        except:
            return timestamp


def _check_if_iso_format(timestamp):
    format_string = "%Y-%m-%dT%H:%M:%S%z"
    try:
        colon = timestamp[-3]
        colonless_timestamp = timestamp
        if colon == ":":
            colonless_timestamp = timestamp[:-3] + timestamp[-2:]
        dt_ob = datetime.datetime.strptime(colonless_timestamp, format_string)
        return True
    except:
        try:
            format_string = "%Y-%m-%dT%H:%M:%S"
            dt_ob = datetime.datetime.strptime(timestamp, format_string)
            return dt_ob
        except:
            return False


def _time_filter(time_extent, ele):
    if time_extent is not None:
        if isinstance(time_extent, datetime.datetime):
            if ele < time_extent:
                return True
            else:
                return False
        elif isinstance(time_extent, list):
            if isinstance(time_extent[0], datetime.datetime) and isinstance(
                time_extent[1], datetime.datetime
            ):
                if time_extent[0] < ele and ele < time_extent[1]:
                    return True
                else:
                    return False

        else:
            return True
    else:
        return True


def _linear_regression(sample_size, date_list, x, y):
    ncoefficient = 2
    if sample_size < ncoefficient:
        _LOGGER.warning(
            "Trend line cannot be drawn. Insufficient points to plot Linear Trend Line"
        )
        return [], []

    AA = _np.empty([sample_size, ncoefficient], dtype=float, order="C")
    BB = _np.empty([sample_size, 1], dtype=float, order="C")
    XX = _np.empty([ncoefficient, 1], dtype=float, order="C")
    for i in range(sample_size):
        n = 0
        AA[i][n] = date_list[i]
        AA[i][n + 1] = 1
        BB[i] = y[i]

    x1 = _np.linalg.lstsq(AA, BB, rcond=None)[0]

    YY = []
    for i in range(sample_size):
        y_temp = x1[0][0] * date_list[i] + x1[1][0]
        YY.append(y_temp)
    return x, YY


def _harmonic_regression(sample_size, date_list, x, y, trend_order):
    PI2_Year = 3.14159265 * 2 / 365.25

    ncoefficient = 2 * (trend_order + 1)
    if sample_size < ncoefficient:
        _LOGGER.warning(
            "Trend line cannot be drawn. Insufficient points to plot Harmonic Trend Line for trend order "
            + str(trend_order)
            + ". Please try specifying a lower trend order."
        )
        return [], []

    AA = _np.empty([sample_size, ncoefficient], dtype=float, order="C")
    BB = _np.empty([sample_size, 1], dtype=float, order="C")
    XX = _np.empty([ncoefficient, 1], dtype=float, order="C")

    for i in range(sample_size):
        n = 0
        AA[i][n] = date_list[i]
        AA[i][n + 1] = 1

        for j in range(1, trend_order + 1):
            AA[i][n + 2 * j] = _np.sin(PI2_Year * j * date_list[i])
            AA[i][n + 2 * j + 1] = _np.cos(PI2_Year * j * date_list[i])

        BB[i] = y[i]

    x1 = _np.linalg.lstsq(AA, BB, rcond=None)[0]
    YY = []
    for i in range(sample_size):
        y_temp = x1[0][0] * date_list[i] + x1[1][0]
        for q in range(2, len(x1), 2):
            y_temp = y_temp + x1[q][0] * _np.sin(
                2 * 3.14159265358979323846 * (q / 2) * date_list[i] / 365.25
            )
            y_temp = y_temp + x1[q + 1][0] * _np.cos(
                2 * 3.14159265358979323846 * (q / 2) * date_list[i] / 365.25
            )
        YY.append(y_temp)
    return x, YY


def _epoch_to_iso(dt):
    import datetime

    try:
        if dt < 0:
            return (
                datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)
                + datetime.timedelta(seconds=(dt / 1000))
            ).isoformat()
        else:
            return datetime.datetime.fromtimestamp(
                dt / 1000, tz=datetime.timezone.utc
            ).isoformat()
    except:
        return dt


def _datetime2ole(date):
    # date = datetime.strptime(date, '%d-%b-%Y')
    import datetime

    OLE_TIME_ZERO = datetime.datetime(1899, 12, 30)
    delta = date - OLE_TIME_ZERO
    return float(delta.days) + (float(delta.seconds) / 86400)


def _ole2datetime(oledt):
    import datetime

    OLE_TIME_ZERO = datetime.datetime(1899, 12, 30, 0, 0, 0)
    try:
        return OLE_TIME_ZERO + datetime.timedelta(days=float(oledt))
    except:
        return datetime.datetime.fromtimestamp(
            oledt / 1000, tz=datetime.timezone.utc
        ).replace(tzinfo=None)


def _iso_to_datetime(timestamp):
    format_string = "%Y-%m-%dT%H:%M:%S%z"
    try:
        colon = timestamp[-3]
        colonless_timestamp = timestamp
        if colon == ":":
            colonless_timestamp = timestamp[:-3] + timestamp[-2:]
        dt_ob = datetime.datetime.strptime(colonless_timestamp, format_string)
        return dt_ob.replace(tzinfo=None)
    except:
        try:
            format_string = "%Y-%m-%dT%H:%M:%S.%f%z"
            dt_ob = datetime.datetime.strptime(colonless_timestamp, format_string)
            return dt_ob.replace(tzinfo=None)
        except:
            try:
                format_string = "%Y-%m-%dT%H:%M:%S"
                dt_ob = datetime.datetime.strptime(timestamp, format_string)
                return dt_ob
            except:
                return timestamp


def _check_if_iso_format(timestamp):
    format_string = "%Y-%m-%dT%H:%M:%S%z"
    try:
        colon = timestamp[-3]
        colonless_timestamp = timestamp
        if colon == ":":
            colonless_timestamp = timestamp[:-3] + timestamp[-2:]
        dt_ob = datetime.datetime.strptime(colonless_timestamp, format_string)
        return True
    except:
        try:
            format_string = "%Y-%m-%dT%H:%M:%S"
            dt_ob = datetime.datetime.strptime(timestamp, format_string)
            return dt_ob
        except:
            return False


def _local_function_template(
    operation_number=None,
    percentile_value=None,
    percentile_interpolation_type=None,
):
    template_dict = {
        "name": "max_rft",
        "description": "A raster function template.",
        "function": {
            "pixelType": "UNKNOWN",
            "name": "Cell Statistics",
            "description": "Calculates a per-cell statistic from multiple rasters.  The available statistics are Majority, Maximum, Mean, Median, Minimum, Minority, Range, Standard Deviation, Sum, and Variety.",
            "type": "LocalFunction",
            "_object_id": 1,
        },
        "arguments": {
            "Rasters": {
                "name": "Rasters",
                "value": {
                    "elements": [],
                    "type": "ArgumentArray",
                    "_object_id": 2,
                },
                "aliases": ["__IsRasterArray__"],
                "isDataset": False,
                "isPublic": False,
                "type": "RasterFunctionVariable",
                "_object_id": 3,
            },
            "Operation": {
                "name": "Operation",
                "value": "",
                "isDataset": False,
                "isPublic": False,
                "type": "RasterFunctionVariable",
                "_object_id": 4,
            },
            "CellsizeType": {
                "name": "CellsizeType",
                "value": 2,
                "isDataset": False,
                "isPublic": False,
                "type": "RasterFunctionVariable",
                "_object_id": 5,
            },
            "ExtentType": {
                "name": "ExtentType",
                "value": 1,
                "isDataset": False,
                "isPublic": False,
                "type": "RasterFunctionVariable",
                "_object_id": 6,
            },
            "ProcessAsMultiband": {
                "name": "ProcessAsMultiband",
                "value": True,
                "isDataset": False,
                "isPublic": False,
                "type": "RasterFunctionVariable",
                "_object_id": 7,
            },
            "MatchVariable": {
                "name": "MatchVariable",
                "value": True,
                "isDataset": False,
                "isPublic": False,
                "type": "RasterFunctionVariable",
                "_object_id": 8,
            },
            "UnionDimension": {
                "name": "UnionDimension",
                "value": False,
                "isDataset": False,
                "isPublic": False,
                "type": "RasterFunctionVariable",
                "_object_id": 9,
            },
            "PercentileValue": {
                "name": "PercentileValue",
                "value": 90,
                "isDataset": False,
                "isPublic": False,
                "type": "RasterFunctionVariable",
                "_object_id": 5,
            },
            "PercentileInterpolationType": {
                "name": "PercentileInterpolationType",
                "value": 1,
                "isDataset": False,
                "isPublic": False,
                "type": "RasterFunctionVariable",
                "_object_id": 6,
            },
            "type": "LocalFunctionArguments",
            "_object_id": 10,
        },
        "functionType": 0,
        "thumbnail": "",
    }
    if operation_number is not None:
        template_dict["arguments"]["Operation"]["value"] = operation_number
    if percentile_value is not None:
        template_dict["arguments"]["PercentileValue"]["value"] = percentile_value
    if percentile_interpolation_type is not None:
        template_dict["arguments"]["PercentileInterpolationType"][
            "value"
        ] = percentile_interpolation_type
    return template_dict


# def _percentile_function_template(
#    ignore_nodata=False, percentile=90, percentile_interpolation_type=False
# ):
#    template_dict = {
#        "name": "Raster Function Template",
#        "description": "A raster function template.",
#        "function": {
#            "pixelType": "UNKNOWN",
#            "name": "Percentile Function",
#            "description": "Compute percentile value across the input rasters.",
#            "type": "PercentileFunction",
#            "_object_id": 1,
#        },
#        "arguments": {
#            "Rasters": {
#                "name": "Rasters",
#                "isDataset": False,
#                "isPublic": False,
#                "type": "RasterFunctionVariable",
#                "_object_id": 2,
#            },
#            "IgnoreNoData": True,
#            "Percentile": 90,
#            "InterpolatePercentile": False,
#            "type": "PercentileFunctionArguments",
#            "_object_id": 3,
#        },
#        "functionType": 0,
#        "type": "RasterFunctionTemplate",
#        "_object_id": 4,
#    }

#    if ignore_nodata is not None:
#        template_dict["arguments"]["IgnoreNoData"] = ignore_nodata

#    if percentile is not None:
#        template_dict["arguments"]["Percentile"] = percentile

#    if percentile_interpolation_type is not None:
#        template_dict["arguments"][
#            "InterpolatePercentile"
#        ] = percentile_interpolation_type

#    return template_dict


def _get_geometry(data):
    if data is None:
        return None

    if isinstance(data, _Geometry):
        return data
    elif isinstance(data, _arcgis.raster.Raster):
        return _Geometry(data.extent)
    elif isinstance(data, _ImageryLayer):
        return _Geometry(data.extent)
    elif isinstance(data, _FeatureLayer):
        return _get_geometry_from_feature_layer(data)
    else:
        return data


def _get_geometry_from_feature_layer(data):
    geo = None
    layer_fset = data.query()
    try:
        for ele in layer_fset.features:
            geo = geo.union(_Geometry(ele.geometry)) if geo else _Geometry(ele.geometry)
    except:
        _LOGGER.warning(
            "Failure while constructing the union of the individual feature geometries"
        )
    return geo


def build_query_string(field_name, operator, field_values):
    operator_map = {
        "equals": "=",
        "less_than": "<",
        "greater_than": ">",
        "not_equals": "<>",
        "not_less_than": ">=",
        "not_greater_than": "<=",
    }

    if operator in operator_map:
        if isinstance(field_values, numbers.Number):
            return field_name + " " + operator_map[operator] + " " + str(field_values)
        elif isinstance(field_values, str):
            return field_name + " " + operator_map[operator] + " '" + field_values + "'"
        else:
            raise TypeError("field_value must be numeric or string")

    elif operator in [
        "starts_with",
        "ends_with",
        "not_starts_with",
        "not_ends_with",
        "contains",
        "not_contains",
    ]:
        if not isinstance(field_values, str):
            raise TypeError("field_value must be string")
        if operator == "starts_with":
            return field_name + " LIKE " + "'" + field_values + "%'"
        elif operator == "ends_with":
            return field_name + " LIKE" + "'%" + field_values + "'"
        elif operator == "not_starts_with":
            return field_name + " NOT LIKE " + "'" + field_values + "%'"
        elif operator == "not_ends_with":
            return field_name + " NOT LIKE " + "'%" + field_values + "'"
        elif operator == "contains":
            return field_name + " LIKE " + "'%" + field_values + "%'"
        elif operator == "not_contains":
            return field_name + " NOT LIKE " + "'%" + field_values + "%'"
    elif operator == "in":
        if not isinstance(field_values, list):
            raise TypeError('field_values must be type list for operator "in"')
        values = "("
        for item in field_values:
            if not (isinstance(item, numbers.Number) or isinstance(item, str)):
                raise TypeError("item in field_values must be numeric or string")
            if values == "(":
                values += "'" + item + "'" if isinstance(item, str) else str(item)
            else:
                values += (
                    ",'" + item + "'" if isinstance(item, str) else "," + str(item)
                )
        values += ")"
        return field_name + " IN " + values
    elif operator == "not_in":
        values = "("
        for item in field_values:
            if not (isinstance(item, numbers.Number) or isinstance(item, str)):
                raise TypeError("item in field_values must be numeric or string")
            if values == "(":
                values += "'" + item + "'" if isinstance(item, str) else str(item)
            else:
                values += (
                    ",'" + item + "'" if isinstance(item, str) else "," + str(item)
                )
        values += ")"
        return field_name + " NOT IN " + values
    else:
        raise ValueError("invalid operator value")


def _generate_direct_access_url(gis=None, expiration=None):
    """helper fn to get the direct access url for azure storage"""
    gis = _arcgis.env.active_gis if gis is None else gis
    url = "%s/sharing/rest/content/users/%s/generateDirectAccessUrl" % (
        gis._portal.url,
        gis.users.me.username,
    )
    params = {"f": "json", "storeType": "rasterStore"}
    if expiration is not None:
        params.update({"expiration": expiration})
    else:
        params.update({"expiration": 1440})
    res = gis._portal.con.post(url, params)
    if isinstance(res, dict):
        if "url" in res.keys():
            return res["url"]
        else:
            raise RuntimeError("Couldn't generate direct access url")
    else:
        raise RuntimeError("Couldn't generate direct access url")


def _print_on_same_line(msg):
    """helper method for printing text on same line"""
    last_msg_length = (
        len(_print_on_same_line.last_msg)
        if hasattr(_print_on_same_line, "last_msg")
        else 0
    )
    print(" " * last_msg_length, end="\r")
    print(msg, end="\r")
    sys.stdout.flush()
    _print_on_same_line.last_msg = msg


def _print_progress_bar(
    iteration,
    total,
    prefix="",
    suffix="",
    decimals=1,
    length=100,
    unit="items",
    fill="█",
):
    """method for displaying a progress bar"""
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filled_length = int(length * iteration // total)
    bar = fill * filled_length + "-" * (length - filled_length)

    _print_on_same_line(
        f"\r{prefix} |{bar}| {percent}% complete ({iteration}/{total} {unit})\t{suffix}"
    )
    if iteration == total:
        print("\n")


def _ra_upload_allowed_extensions():
    """
    returns a list of valid file upload extensions supported for creating
    hosted imagery layers from raster datasets on Enterprise/AGOL
    """
    return (
        "1b,5gud,a11,a12,a13,a14,a15,a16,a17,a18,a19,a1a,a1b,a1c,a1d,a1e,"
        "a1f,a1g,a1h,a1j,a21,a22,a23,a24,a25,a26,a27,a28,a29,a2a,a2b,a2c,"
        "a2d,a2e,a2f,a2g,a2h,a2j,a31,a32,a33,a34,a35,a36,a37,a38,a39,a3a,"
        "a3b,a3c,a3d,a3e,a3f,a3g,a3h,a3j,a41,a42,a43,a44,a45,a46,a47,a48,"
        "a49,a4a,a4b,a4c,a4d,a4e,a4f,a4g,a4h,a4j,adf,ads,afr,asc,at1,at2,"
        "at3,at4,at5,at6,at7,at8,at9,ata,atb,atc,atd,ate,atf,atg,ath,atj,"
        "att,aux,avg,bag,bil,bin,bip,blw,blx,bmp,bpw,bqw,bsq,bt,bundle,"
        "bundlx,c11,c12,c13,c14,c15,c16,c17,c18,c19,c1a,c1b,c1c,c1d,c1e,"
        "c1f,c1g,c1h,c1j,c21,c22,c23,c24,c25,c26,c27,c28,c29,c2a,c2b,c2c,"
        "c2d,c2e,c2f,c2g,c2h,c2j,c41,c42,c43,c44,c45,c46,c47,c48,c49,c4a,"
        "c4b,c4c,c4d,c4e,c4f,c4g,c4h,c4j,c51,c52,c53,c54,c55,c56,c57,c58,"
        "c59,c5a,c5b,c5c,c5d,c5e,c5f,c5g,c5h,c5j,c61,c62,c63,c64,c65,c66,"
        "c67,c68,c69,c6a,c6b,c6c,c6d,c6e,c6f,c6g,c6h,c6j,c71,c72,c73,c74,"
        "c75,c76,c77,c78,c79,c7a,c7b,c7c,c7d,c7e,c7f,c7g,c7h,c7j,c81,c82,"
        "c83,c84,c85,c86,c87,c88,c89,c8a,c8b,c8c,c8d,c8e,c8f,c8g,c8h,c8j,"
        "c91,c92,c93,c94,c95,c96,c97,c98,c99,c9a,c9b,c9c,c9d,c9e,c9f,c9g,"
        "c9h,c9j,ca1,ca2,ca3,ca4,ca5,ca6,ca7,ca8,ca9,caa,cab,cac,cad,cae,"
        "caf,cag,cah,caj,cb1,cb2,cb3,cb4,cb5,cb6,cb7,cb8,cb9,cba,cbb,cbc,"
        "cbd,cbe,cbf,cbg,cbh,cbj,cc1,cc2,cc3,cc4,cc5,cc6,cc7,cc8,cc9,cca,"
        "ccb,ccc,ccd,cce,ccf,ccg,cch,ccj,cd1,cd2,cd3,cd4,cd5,cd6,cd7,cd8,"
        "cd9,cda,cdb,cdc,cdd,cde,cdf,cdg,cdh,cdi,cdj,ce1,ce2,ce3,ce4,ce5,"
        "ce6,ce7,ce8,ce9,cea,ceb,cec,ced,cee,cef,ceg,ceh,cej,cf1,cf2,cf3,"
        "cf4,cf5,cf6,cf7,cf8,cf9,cfa,cfb,cfc,cfd,cfe,cff,cfg,cfh,cfj,cg1,"
        "cg2,cg3,cg4,cg5,cg6,cg7,cg8,cg9,cga,cgb,cgc,cgd,cge,cgf,cgg,cgh,"
        "cgj,ch1,ch2,ch3,ch4,ch5,ch6,ch7,ch8,ch9,cha,chb,chc,chd,che,chf,"
        "chg,chh,chj,cit,cj1,cj2,cj3,cj4,cj5,cj6,cj7,cj8,cj9,cja,cjb,cjc,"
        "cjd,cje,cjf,cjg,cjh,cjj,ck1,ck2,ck3,ck4,ck5,ck6,ck7,ck8,ck9,cka,"
        "ckb,ckc,ckd,cke,ckf,ckg,ckh,ckj,cl1,cl2,cl3,cl4,cl5,cl6,cl7,cl8,"
        "cl9,cla,clb,clc,cld,cle,clf,clg,clh,clj,clr,cm1,cm2,cm3,cm4,cm5,"
        "cm6,cm7,cm8,cm9,cma,cmb,cmc,cmd,cme,cmf,cmg,cmh,cmj,cn1,cn2,cn3,"
        "cn4,cn5,cn6,cn7,cn8,cn9,cna,cnb,cnc,cnd,cne,cnf,cng,cnh,cnj,co1,"
        "co2,co3,co4,co5,co6,co7,co8,co9,coa,cob,coc,cod,coe,cof,cog,coh,"
        "coj,cos,cot,cp1,cp2,cp3,cp4,cp5,cp6,cp7,cp8,cp9,cpa,cpb,cpc,cpd,"
        "cpe,cpf,cpg,cph,cpj,cq1,cq2,cq3,cq4,cq5,cq6,cq7,cq8,cq9,cqa,cqb,"
        "cqc,cqd,cqe,cqf,cqg,cqh,cqj,cr1,cr2,cr3,cr4,cr5,cr6,cr7,cr8,cr9,"
        "cra,crb,crc,crd,cre,crf,crg,crh,crj,cs1,cs2,cs3,cs4,cs5,cs6,cs7,"
        "cs8,cs9,csa,csb,csc,csd,cse,csf,csg,csh,csj,ct1,ct2,ct3,ct4,ct5,"
        "ct6,ct7,ct8,ct9,cta,ctb,ctc,ctd,cte,ctf,ctg,cth,ctj,cub,dat,dbf,"
        "ddf,dem,dim,dt0,dt1,dt2,elas,eph,ers,f11,f12,f13,f14,f15,f16,f17,"
        "f18,f19,f1a,f1b,f1c,f1d,f1e,f1f,f1g,f1h,f1j,f21,f22,f23,f24,f25,"
        "f26,f27,f28,f29,f2a,f2b,f2c,f2d,f2e,f2f,f2g,f2h,f2j,f31,f32,f33,"
        "f34,f35,f36,f37,f38,f39,f3a,f3b,f3c,f3d,f3e,f3f,f3g,f3h,f3j,f41,"
        "f42,f43,f44,f45,f46,f47,f48,f49,f4a,f4b,f4c,f4d,f4e,f4f,f4g,f4h,"
        "f4j,f51,f52,f53,f54,f55,f56,f57,f58,f59,f5a,f5b,f5c,f5d,f5e,f5f,"
        "f5g,f5h,f5j,fit,flt,fst,gc,geo,gff,gif,gis,gn1,gn2,gn3,gn4,gn7,gn9,"
        "gna,gnb,gnc,gnd,gng,gnj,gr2,grb,grb2,grc,grd,grib,grib2,gtx,gxf,h1,"
        "h4,h5,ha1,ha1,ha2,ha3,ha4,ha5,ha6,ha7,ha8,ha9,haa,hab,hac,had,hae,"
        "haf,hag,hah,haj,hdf,hdf4,hdf5,hdr,he4,he5,hf2,hgt,hr1,hr2,hr3,hr4,"
        "hr5,hr6,hr7,hr8,i1,i11,i12,i13,i14,i15,i16,i17,i18,i19,i1a,i1b,i1c,"
        "i1d,i1e,i1f,i1g,i1h,i1j,i2,i21,i22,i23,i24,i25,i26,i27,i28,i29,i2a,"
        "i2b,i2c,i2d,i2e,i2f,i2g,i2h,i2j,i3,i31,i32,i33,i34,i35,i36,i37,i38,"
        "i39,i3a,i3b,i3c,i3d,i3e,i3f,i3g,i3h,i3j,i4,i41,i42,i43,i44,i45,i46,"
        "i47,i48,i49,i4a,i4b,i4c,i4d,i4e,i4f,i4g,i4h,i4j,i5,i51,i52,i53,i54,"
        "i55,i56,i57,i58,i59,i5a,i5b,i5c,i5d,i5e,i5f,i5g,i5h,i5j,i6,i7,i8,i9,"
        "idx,ige,imd,img,iv1,iv2,iv3,iv4,iv5,iv6,iv7,iv8,iv9,iva,ivb,ivc,ivd,"
        "ive,ivf,ivg,ivh,ivj,j2c,j2k,ja1,ja2,ja3,ja4,ja5,ja6,ja7,ja8,ja9,jaa,"
        "jab,jac,jad,jae,jaf,jag,jah,jaj,jg1,jg2,jg3,jg4,jg5,jg6,jg7,jg8,jg9,"
        "jga,jgb,jgc,jgd,jge,jgf,jgg,jgh,jgj,jgw,jn1,jn2,jn3,jn4,jn5,jn6,jn7,"
        "jn8,jn9,jna,jnb,jnc,jnd,jne,jnf,jng,jnh,jnj,jo1,jo2,jo3,jo4,jo5,jo6,"
        "jo7,jo8,jo9,joa,job,joc,jod,joe,jof,jog,joh,joj,jp2,jpc,jpg,jpw,jpx,"
        "jr1,jr2,jr3,jr4,jr5,jr6,jr7,jr8,jr9,jra,jrb,jrc,jrd,jre,jrf,jrg,jrh,"
        "jrj,json,kap,l11,l12,l13,l14,l15,l16,l17,l18,l19,l1a,l1b,l1c,l1d,l1e,"
        "l1f,l1g,l1h,l1j,l21,l22,l23,l24,l25,l26,l27,l28,l29,l2a,l2b,l2c,l2d,"
        "l2e,l2f,l2g,l2h,l2j,l31,l32,l33,l34,l35,l36,l37,l38,l39,l3a,l3b,l3c,"
        "l3d,l3e,l3f,l3g,l3h,l3j,l41,l42,l43,l44,l45,l46,l47,l48,l49,l4a,l4b,"
        "l4c,l4d,l4e,l4f,l4g,l4h,l4j,l51,l52,l53,l54,l55,l56,l57,l58,l59,l5a,"
        "l5b,l5c,l5d,l5e,l5f,l5g,l5h,l5j,lan,las,lbl,lf1,lf2,lf3,lf4,lf5,lf6,"
        "lf7,lf8,lf9,lfa,lfb,lfc,lfd,lfe,lff,lfg,lfh,lfj,lgg,ln1,ln2,ln3,ln4,"
        "ln5,ln6,ln7,ln8,ln9,lna,lnb,lnc,lnd,lne,lnf,lng,lnh,lnj,lrc,m11,m12,"
        "m13,m14,m15,m16,m17,m18,m19,m1a,m1b,m1c,m1d,m1e,m1f,m1g,m1h,m1j,m21,"
        "m22,m23,m24,m25,m26,m27,m28,m29,m2a,m2b,m2c,m2d,m2e,m2f,m2g,m2h,m2j,"
        "map,max,memory,met,mi1,mi2,mi3,mi4,mi5,mi6,mi7,mi8,mi9,mia,mib,mic,"
        "mid,mie,mif,mig,mih,mij,min,mm1,mm2,mm3,mm4,mm5,mm6,mm7,mm8,mm9,mma,"
        "mmb,mmc,mmd,mme,mmf,mmg,mmh,mmj,mpl,mpr,mrf,mtl,n1,nc,nc4,nes,nsf,ntf,"
        "oa1,oa2,oa3,oa4,oa5,oa6,oa7,oa8,oa9,oaa,oab,oac,oad,oae,oaf,oag,oah,"
        "oaj,oh1,oh2,oh3,oh4,oh5,oh6,oh7,oh8,oh9,oha,ohb,ohc,ohd,ohe,ohf,ohg,"
        "ohh,ohj,on1,on2,on3,on4,on5,on6,on7,on8,on9,ona,onb,onc,ond,one,onf,"
        "ong,onh,onj,ovr,ow1,ow2,ow3,ow4,ow5,ow6,ow7,ow8,ow9,owa,owb,owc,owd,"
        "owe,owf,owg,owh,owj,paux,pbm,pgm,pgw,pix,png,ppm,prj,pro,properties,"
        "psi,pvl,r0,raw,rpb,rpc,rrd,rst,rv1,rv2,rv3,rv4,rv5,rv6,rv7,rv8,rv9,"
        "rva,rvb,rvc,rvd,rve,rvf,rvg,rvh,rvj,sdat,sdw,sid,sta,stk,sv,tc1,tc2,"
        "tc3,tc4,tc5,tc6,tc7,tc8,tc9,tca,tcb,tcc,tcd,tce,tcf,tcg,tch,tcj,ter,"
        "tf1,tf2,tf3,tf4,tf5,tf6,tf7,tf8,tf9,tfa,tfb,tfc,tfd,tfe,tff,tfg,tfh,"
        "tfj,tfrd,tfw,tif,tiff,til,tl1,tl2,tl3,tl4,tl5,tl6,tl7,tl8,tl9,tla,tlb,"
        "tlc,tld,tle,tlf,tlg,tlh,tlj,tn1,tn2,tn3,tn4,tn5,tn6,tn7,tn8,tn9,tna,"
        "tnb,tnc,tnd,tne,tnf,tng,tnh,tnj,toc,tp1,tp2,tp3,tp4,tp5,tp6,tp7,tp8,"
        "tp9,tpa,tpb,tpc,tpd,tpe,tpf,tpg,tph,tpj,tq1,tq2,tq3,tq4,tq5,tq6,tq7,"
        "tq8,tq9,tqa,tqb,tqc,tqd,tqe,tqf,tqg,tqh,tqj,tr1,tr2,tr3,tr4,tr5,tr6,"
        "tr7,tr8,tr9,tra,trb,trc,trd,tre,trf,trg,trh,trj,trl,tt1,tt2,tt3,tt4,"
        "tt5,tt6,tt7,tt8,tt9,tta,ttb,ttc,ttd,tte,ttf,ttg,tth,ttj,txt,ul1,ul2,"
        "ul3,ul4,ul5,ul6,ul7,ul8,ul9,ula,ulb,ulc,uld,ule,ulf,ulg,ulh,ulj,vh1,"
        "vh2,vh3,vh4,vh5,vh6,vh7,vh8,vh9,vha,vhb,vhc,vhd,vhe,vhf,vhg,vhh,vhj,"
        "view,vn1,vn2,vn3,vn4,vn5,vn6,vn7,vn8,vn9,vna,vnb,vnc,vnd,vne,vnf,vng,"
        "vnh,vnj,vrt,vt1,vt2,vt3,vt4,vt5,vt6,vt7,vt8,vt9,vta,vtb,vtc,vtd,vte,"
        "vtf,vtg,vth,vtj,wo,xml,xpm,xyz,gdb,pjg,pzp,ppg".split(",")
    )


def _is_primary_file(file):
    """Returns the file path if it is a primary file, otherwise return None."""

    # CRF folder check
    if file.endswith(".bundle") and "/_alllayers" in file:
        return file[: file.rfind("/_alllayers")]

    # Common raster dataset formats check
    if file.lower().endswith(
        (
            ".tiff",
            ".tif",
            ".mrf",
            ".img",
            ".jp2",
            ".jpx",
            ".j2k",
            ".sid",
            ".ntf",
            ".nsf",
            ".hdf",
            ".hdf4",
            ".hdf5",
            ".h4",
            ".h5",
            ".he4",
            ".he5",
            ".grib",
            ".grb",
            ".grib2",
            ".grb2",
            ".bin",
            ".dat",
            ".nc",
            ".nc4",
        )
    ):
        return file
    return


class _ImageryUploaderAGOL:
    """helper class for concurrently uploading multiple files to user's rasterstore on AGOL"""

    def __init__(
        self,
        file_list,
        container,
        auto_renew,
        upload_properties,
        task,
        raster_type,
        gis,
    ):
        from azure.storage.blob import ContainerClient
        from azure.core.exceptions import (
            ClientAuthenticationError,
            ServiceResponseError,
            ServiceRequestError,
        )

        self.ContainerClient = ContainerClient
        (
            self.ClientAuthenticationError,
            self.ServiceResponseError,
            self.ServiceRequestError,
        ) = (
            ClientAuthenticationError,
            ServiceResponseError,
            ServiceRequestError,
        )

        self.file_list = file_list
        self.container = container
        self.auto_renew = auto_renew
        self.task = task
        self.raster_type = raster_type
        self.gis = gis
        self.all_files = []
        self.mosaic_data_info = []
        self.primary_files = []
        self.single_primary_file = (
            False
            if self.raster_type != "Raster Dataset"
            or (
                len(self.file_list) > 1
                and any(item["is_dir"] for item in self.file_list)
            )
            else True
        )
        for i, d in enumerate(file_list):
            self.all_files.extend([(f, i) for f in d["files_list"]])
        self.url_list = []
        use_defaults = False
        if upload_properties is None or not isinstance(upload_properties, dict):
            upload_properties_lower = {
                "maxuploadconcurrency": 6,
                "maxworkerthreads": None,
                "displayprogress": False,
            }
            use_defaults = True
        else:
            upload_properties_lower = {
                k.lower(): v for k, v in upload_properties.items()
            }

        if "maxuploadconcurrency" in upload_properties_lower and (
            isinstance(upload_properties_lower["maxuploadconcurrency"], int)
            or upload_properties_lower["maxuploadconcurrency"] is None
        ):
            self.max_upload_concurrency = upload_properties_lower[
                "maxuploadconcurrency"
            ]
        else:
            self.max_upload_concurrency = 6

        if "maxworkerthreads" in upload_properties_lower and (
            isinstance(upload_properties_lower["maxworkerthreads"], int)
            or upload_properties_lower["maxworkerthreads"] is None
        ):
            self.max_worker_threads = upload_properties_lower["maxworkerthreads"]
        else:
            self.max_worker_threads = None

        display_flag = _arcgis.env.verbose
        if "displayprogress" in upload_properties_lower and isinstance(
            upload_properties_lower["displayprogress"], bool
        ):
            self.display_progress = upload_properties_lower["displayprogress"]
            if not self.display_progress and not use_defaults:
                display_flag = False
        else:
            self.display_progress = False

        if display_flag:
            self.display_progress = True

    def upload_file(self, file_item):
        """method to upload single file"""

        file_name, i = file_item
        prefix = self.file_list[i]["prefix"]
        current_time_str = prefix[:-1][8:]
        is_dir = self.file_list[i]["is_dir"]
        data_for_md = self.file_list[i]["data_for_md"]
        if is_dir:
            folder_path = self.file_list[i]["file_name"]
            root = os.path.dirname(file_name)
            basename_len = self.file_list[i]["basename_len"]
            blobname = prefix + (root + "/" + os.path.basename(file_name))[
                basename_len + 1 :
            ].replace(os.sep, "/")
        else:
            blobname = prefix + os.path.basename(file_name).replace(os.sep, "/")

        while True:
            try:
                blob = self.container.get_blob_client(blobname)

                with open(file_name, "rb") as data:
                    blob.upload_blob(
                        data,
                        blob_type="BlockBlob",
                        max_concurrency=self.max_upload_concurrency,
                    )

                url = blob.url.split("?", 1)[0]
                url_suffix = (
                    "arcgis.com" if "arcgis.com" in url else "blob.core.windows.net"
                )

                if is_dir:
                    if self.file_list[i]["single_image"]:
                        self.url_list.append(url)
                    else:
                        if url != "":
                            if data_for_md:
                                source = folder_path
                                if self.task == "CreateImageCollection":
                                    target = os.path.basename(source)
                                else:
                                    folder_match = quote(
                                        prefix + os.path.basename(source)
                                    )
                                    folder_url = url[
                                        0 : url.find(folder_match) + len(folder_match)
                                    ]
                                    target = unquote(
                                        folder_url.replace(
                                            folder_url[
                                                0 : folder_url.find(url_suffix)
                                                + len(url_suffix)
                                            ],
                                            "/vsiaz",
                                        )
                                    )

                                data_path = {
                                    "source": source,
                                    "target": target,
                                }
                                if data_path not in self.mosaic_data_info:
                                    self.mosaic_data_info.append(data_path)

                            if self.single_primary_file:
                                primary_file = _is_primary_file(url)
                                if primary_file:
                                    if primary_file not in self.primary_files:
                                        self.primary_files.append(primary_file)
                            url = url[
                                0 : url.find(current_time_str) + len(current_time_str)
                            ]
                            if url not in self.url_list:
                                self.url_list.append(url)
                else:
                    if data_for_md:
                        if self.task == "CreateImageCollection":
                            data_path = os.path.dirname(file_name)
                        else:
                            source = os.path.dirname(file_name)
                            target = os.path.dirname(
                                url.replace(
                                    url[0 : url.find(url_suffix) + len(url_suffix)],
                                    "/vsiaz",
                                )
                            )
                            data_path = {"source": source, "target": target}

                        if data_path not in self.mosaic_data_info:
                            self.mosaic_data_info.append(data_path)

                    if self.single_primary_file:
                        primary_file = _is_primary_file(url)
                        if primary_file:
                            if primary_file not in self.primary_files:
                                self.primary_files.append(primary_file)

                    if (
                        url not in self.url_list
                        and os.path.dirname(url) not in self.url_list
                    ):
                        self.url_list.append(url)

                break
            except (
                self.ClientAuthenticationError,
                self.ServiceResponseError,
                self.ServiceRequestError,
            ) as err:
                if self.auto_renew:
                    sas_url = _generate_direct_access_url(self.gis)
                    self.container = self.ContainerClient.from_container_url(sas_url)
                    continue
                else:
                    raise
            except Exception as err:
                raise err

        return file_name

    def upload_all_files(self):
        """method to upload multiple files concurrently"""
        result, mosaic_data_info = self.run(self.all_files)
        return result, mosaic_data_info

    def run(self, all_files):
        """helper method for creating a thread pool for uploading multiple files"""
        import concurrent.futures as _cf

        if self.display_progress:
            l = len(all_files)
            start = time.time()
            _print_progress_bar(0, l, prefix=" ", suffix=" ", length=30, unit="files")

        with _cf.ThreadPoolExecutor(self.max_worker_threads) as executor:
            futures = [executor.submit(self.upload_file, arg) for arg in all_files]

            if self.display_progress:
                current = 0
                for future in _cf.as_completed(futures):
                    res = future.result()
                    current += 1
                    end = time.time()
                    elapsed_time = str(datetime.timedelta(seconds=end - start))
                    time_comp = ":" if current == l else ">"
                    _print_progress_bar(
                        current,
                        l,
                        prefix=f"{os.path.basename(res)} uploaded",
                        suffix=f"[time elapsed{time_comp} {elapsed_time}]",
                        length=30,
                        unit="files",
                    )

        if len(self.primary_files) == 1:
            self.url_list[:] = self.primary_files
        return self.url_list, self.mosaic_data_info


def _upload_imagery_agol(
    files,
    gis=None,
    direct_access_url=None,
    auto_renew=True,
    upload_properties=None,
    single_image=False,
    raster_type=None,
    task=None,
):
    """uploads imagery to user's rasterstore on AGOL and returns the list of urls"""

    try:
        from azure.storage.blob import ContainerClient
        from azure.core.exceptions import (
            ClientAuthenticationError,
            ServiceResponseError,
            ServiceRequestError,
        )
    except:
        _LOGGER.warning(
            "Install Azure library packages for Python."
            + "(Azure SDK for Python - azure-storage-blob: 12.1<= version <=12.17)"
            + "\n(https://docs.microsoft.com/en-us/azure/storage/blobs/storage-quickstart-blobs-python#install-the-package)"
        )
    gis = _arcgis.env.active_gis if gis is None else gis
    if direct_access_url is None:
        sas_url = _generate_direct_access_url(gis)
    else:
        sas_url = direct_access_url
    container = ContainerClient.from_container_url(sas_url)
    if not isinstance(files, list):
        files = [files]

    file_list = []
    time_list = []
    allowed_extensions = _ra_upload_allowed_extensions()

    is_data_for_md = False
    if (isinstance(raster_type, str)) and raster_type == "mosaic_dataset":
        is_data_for_md = True
        current_time = int(time.time())

    all_files = True
    for file in files:
        to_upload = True
        file_dict = {}
        if not is_data_for_md:
            current_time = int(time.time())
            while current_time in time_list:
                current_time += 1
            time_list.append(current_time)
        file_dict["prefix"] = "_images/" + str(current_time) + "/"
        file_dict["file_name"] = file
        file_dict["single_image"] = single_image
        if is_data_for_md:
            file_dict["data_for_md"] = True
        else:
            file_dict["data_for_md"] = False
        if os.path.exists(file):
            if os.path.isdir(file):
                all_files = False
                file_dict["is_dir"] = True
                file_dict["basename_len"] = len(os.path.dirname(file))
                if not ".gdb" in file:
                    file_dict["files_list"] = [
                        os.path.join(root, f)
                        for root, d_names, f_names in os.walk(file)
                        for f in f_names
                        if os.path.splitext(f)[1][1:].lower() in allowed_extensions
                    ]
                else:
                    file_dict["files_list"] = [
                        os.path.join(root, f)
                        for root, d_names, f_names in os.walk(file)
                        for f in f_names
                    ]

                if len(file_dict["files_list"]) == 0:
                    to_upload = False
            else:
                file_dict["is_dir"] = False
                if os.path.splitext(file)[1][1:].lower() in allowed_extensions:
                    file_dict["files_list"] = [file]
                else:
                    to_upload = False

            if to_upload:
                file_list.append(file_dict)
    if len(file_list) == 0:
        raise RuntimeError("No supported files to upload")
    if all_files:
        for file in file_list:
            file["prefix"] = file_list[0]["prefix"]

    uploader = _ImageryUploaderAGOL(
        file_list,
        container,
        auto_renew,
        upload_properties,
        task,
        raster_type,
        gis,
    )
    mosaic_data_info = []
    url_list, mosaic_data_info = uploader.upload_all_files()
    if is_data_for_md:
        return url_list, mosaic_data_info
    return url_list


def _upload_imagery_enterprise(files, raster_type_name=None, gis=None):
    """uploads a file to the image layer to enterprise and returns the item id"""

    ra_url = gis.properties.helperServices["rasterAnalytics"]["url"]
    url = "%s/uploads/upload" % ra_url
    params = {"f": "json"}

    if not isinstance(files, list):
        files = [files]

    item_ids_list = []
    res = {}

    append_path = False
    for file in files:
        item_id_dict = {}
        if os.path.exists(file):
            if os.path.isdir(file):
                if file.endswith(".crf") or raster_type_name != "Raster Dataset":
                    append_path = True
                elif not file.endswith(".crf") or raster_type_name == "Raster Dataset":
                    for dir_ele in [x[0] for x in os.walk(file)]:
                        if dir_ele.endswith(".crf"):
                            append_path = True  # case when parent of the crf folder is specified and raster type is specified as Raster Dataset, we need to append path
                folder = os.path.basename(file)
                basename_len = len(os.path.dirname(file))
                for root, d_names, f_names in os.walk(file):
                    for f in f_names:
                        fp = os.path.join(root, f)
                        path = ("/" + root + "/" + f)[basename_len + 1 :].replace(
                            os.sep, "/"
                        )
                        item_id = None
                        try:
                            item_id = _upload(path=fp, gis=gis)
                        except Exception as e:
                            if "(Error Code: 403)" in str(e):
                                pass
                            else:
                                _LOGGER.warning("file: " + str(fp) + " " + str(e))

                        if item_id is not None:
                            if append_path:
                                item_id_dict = {
                                    "itemId": item_id,
                                    "path": path,
                                }
                                item_ids_list.append(item_id_dict)
                                item_id_dict = {}
                            else:
                                item_ids_list.append(item_id)

            else:
                files_param = {"file": file}
                item_id = None
                try:
                    item_id = _upload(path=file, gis=gis)
                except Exception as e:
                    _LOGGER.warning("file: " + str(file) + " " + str(e))
                if item_id is not None:
                    item_ids_list.append(item_id)

    return item_ids_list

    # ----------------------------------------------------------------------


def _upload(path, description=None, gis=None):
    """
    The ``upload`` method uploads a new item to the server.

    .. note::
        Once the operation is completed successfully, item id of the uploaded item is returned.

    ===============     ====================================================================
    **Parameter**        **Description**
    ---------------     --------------------------------------------------------------------
    path                Optional string. Filepath of the file to upload.
    ---------------     --------------------------------------------------------------------
    description         Optional string. Descriptive text for the uploaded item.
    ===============     ====================================================================

    :return: Item id of uploaded item

    """
    ra_url = gis.properties.helperServices["rasterAnalytics"]["url"]
    if (os.path.getsize(path)) < 1000000000:
        url = ra_url + "/uploads/upload"
        params = {
            "f": "json",
            "filename": os.path.basename(path),
            "overwrite": True,
        }
        files = {}
        files["file"] = path
        if description:
            params["description"] = description
        res = gis._con.post(path=url, postdata=params, files=files, timeout=None)
        if "error" in res:
            raise Exception(res)
        else:
            return res["item"]["itemID"]
    else:
        file_path = path
        item_id = _register_upload(file_path, gis=gis)
        _upload_by_parts(item_id, file_path, gis=gis)
        return _commit_upload(item_id, gis=gis)


# ----------------------------------------------------------------------
def _register_upload(file_path, gis=None):
    """returns the itemid for the upload by parts logic"""
    ra_url = gis.properties.helperServices["rasterAnalytics"]["url"]
    r_url = "%s/uploads/register" % ra_url
    params = {"f": "json", "itemName": os.path.basename(file_path)}
    reg_res = gis._con.post(r_url, params, timeout=None)
    if "item" in reg_res and "itemID" in reg_res["item"]:
        return reg_res["item"]["itemID"]
    return None


# ----------------------------------------------------------------------
def _upload_by_parts(item_id, file_path, gis=None):
    """loads a file for attachmens by parts"""
    import mmap, tempfile

    ra_url = gis.properties.helperServices["rasterAnalytics"]["url"]
    b_url = "%s/uploads/%s" % (ra_url, item_id)
    upload_part_url = "%s/uploadPart" % b_url
    params = {"f": "json"}
    with open(file_path, "rb") as f:
        mm = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
        size = 100000000
        steps = int(os.fstat(f.fileno()).st_size / size)
        if os.fstat(f.fileno()).st_size % size > 0:
            steps += 1
        for i in range(steps):
            files = {}
            tempFile = os.path.join(tempfile.gettempdir(), "split.part%s" % i)
            if os.path.isfile(tempFile):
                os.remove(tempFile)
            with open(tempFile, "wb") as writer:
                writer.write(mm.read(size))
                writer.flush()
                writer.close()
            del writer
            files["file"] = tempFile
            params["partId"] = i + 1
            res = gis._con.post(
                upload_part_url, postdata=params, files=files, timeout=None
            )
            if "error" in res:
                raise Exception(res)
            os.remove(tempFile)
            del files
        del mm
    return True


# ----------------------------------------------------------------------
def _commit_upload(item_id, gis=None):
    """commits an upload by parts upload"""

    ra_url = gis.properties.helperServices["rasterAnalytics"]["url"]
    b_url = "%s/uploads/%s" % (ra_url, item_id)
    commit_part_url = "%s/commit" % b_url
    params = {"f": "json", "parts": _uploaded_parts(itemid=item_id, gis=gis)}
    res = gis._con.post(commit_part_url, params, timeout=None)
    if "error" in res:
        raise Exception(res)
    else:
        return res["item"]["itemID"]


# ----------------------------------------------------------------------
def _uploaded_parts(itemid, gis=None):
    """
    returns the parts uploaded for a given item

    ==================   ==============================================
    Arguments           Description
    ------------------   ----------------------------------------------
    itemid               required string. Id of the uploaded by parts item.
    ==================   ==============================================

    """
    ra_url = gis.properties.helperServices["rasterAnalytics"]["url"]
    url = ra_url + "/uploads/%s/parts" % itemid
    params = {"f": "json"}
    res = gis._con.get(url, params)
    return ",".join(res["parts"])


def _get_extent(extdict=None):
    """
    This method is used to convert the JSON presentation of extent (with spatial reference)
    to arcpy.Extent object, so that it can be set to the GP environment.
    :param context: context parameter contains output spatial reference info
    :return geometry object and geometry coordinate
    """
    try:
        import arcpy
    except:
        return None, None
    outext = arcpy.Extent
    extsr = ""
    try:
        if extdict is None:
            return outext, extsr
        # Note: creating geometry directly from envelope JSON gave me a _passthrough
        # which does not provide a extent object.
        if (
            "xmin" in extdict
            and "xmax" in extdict
            and "ymin" in extdict
            and "ymax" in extdict
        ):
            xmin = extdict["xmin"]
            ymin = extdict["ymin"]
            xmax = extdict["xmax"]
            ymax = extdict["ymax"]
            extjson = {
                "rings": [
                    [
                        [xmin, ymin],
                        [xmin, ymax],
                        [xmax, ymax],
                        [xmax, ymin],
                        [xmin, ymin],
                    ]
                ]
            }
            if "spatialReference" in extdict:
                srdict = extdict["spatialReference"]
                extjson.update({"spatialReference": srdict})
                extsr = srdict

            polygon = arcpy.AsShape(extjson, True)
            outext = polygon.extent
        return outext, extsr
    except:
        return outext, extsr


def _get_stac_api_search_items(
    api_search_endpoint, query, request_method, request_params, get_all_items
):
    """
    This method is used to retrieve all the STAC Items from a search query.
    :param api_search_endpoint: URL of the STAC API (/search) endpoint. The STAC API where
                                the search needs to be performed.
    :param query: The GET/POST request query dictionary that can be used to query a
                  STAC API's search endpoint.
    :param request_method: The HTTP request method used with the STAC API for
                           making the search ("GET" or "POST").
    :param request params: requests.get()/post() method parameters used for
                           the STAC API search request (specified in dictionary format).
    :param get_all_items: Boolean speciying whether to return all the items (retrieving
                          them from all the pages) or just the items from the first
                          page of matches.
    :return list (of STAC Item dictionaries)
    """

    all_items = []
    more_items = True

    while more_items:
        if request_method.upper() == "GET":
            data = _requests.get(api_search_endpoint, params=query, **request_params)
        else:
            data = _requests.post(api_search_endpoint, json=query, **request_params)

        if data.status_code != 200 or data.headers.get("content-type") not in [
            "application/json",
            "application/geo+json",
            "application/json;charset=utf-8",
            "application/geo+json; charset=utf-8",
        ]:
            raise RuntimeError(
                f"Invalid Response: Please verify that the specified query is correct-\n{data.text}"
            )

        json_data = data.json()
        if "type" not in json_data or json_data["type"] != "FeatureCollection":
            raise RuntimeError(
                f"Invalid JSON Response from the STAC API: Please verify that the specified query is correct-\n{json_data}"
            )

        json_data = data.json()
        items = json_data["features"]
        if not get_all_items:
            return items
        all_items.extend(items)
        next_request = next(
            (link for link in json_data["links"] if link["rel"] == "next"), None
        )
        if next_request:
            query = (
                urlparse(next_request["href"]).query
                if request_method.upper() == "GET"
                else dict(query, **next_request["body"])
            )
        else:
            more_items = False

    return all_items


def _get_stac_metadata_file(item):
    """
    This method is used to retrieve the metadata file of a valid STAC item.
    :param item: input STAC Item (JSON dictionary)
    :return string (URL of the STAC Item metadata file)
    """

    planetary_computer_map = {
        **dict.fromkeys(
            [
                "3dep-seamless",
                "3dep-lidar-dsm",
                "cop-dem-glo-30",
                "cop-dem-glo-90",
                "3dep-lidar-hag",
                "3dep-lidar-intensity",
                "3dep-lidar-pointsourceid",
                "noaa-c-cap",
                "3dep-lidar-returns",
                "3dep-lidar-dtm-native",
                "3dep-lidar-classification",
                "3dep-lidar-dtm",
                "gap",
                "alos-dem",
                "io-lulc",
                "drcog-lulc",
                "chesapeake-lc-7",
                "chesapeake-lc-13",
                "chesapeake-lu",
                "io-lulc-9-class",
                "io-biodiversity",
                "ecmwf-forecast",
            ],
            "data",
        ),
        **dict.fromkeys(
            [
                "sentinel-1-rtc",
                "hgb",
                "gnatsgo-rasters",
                "mobi",
                "chloris-biomass",
                "jrc-gsw",
                "hrea",
                "noaa-nclimgrid-monthly",
                "usda-cdl",
                "esa-cci-lc",
                "noaa-climate-normals-gridded",
                "noaa-cdr-sea-surface-temperature-whoi",
                "noaa-cdr-ocean-heat-content",
                "esa-worldcover",
            ],
            "All COGs",
        ),
        **dict.fromkeys(
            [
                "daymet-annual-pr",
                "daymet-daily-hi",
                "gridmet",
                "daymet-annual-na",
                "daymet-monthly-na",
                "daymet-annual-hi",
                "daymet-monthly-hi",
                "daymet-monthly-pr",
                "terraclimate",
                "daymet-daily-pr",
                "daymet-daily-na",
            ],
            "zarr-https",
        ),
        **dict.fromkeys(
            [
                "sentinel-1-grd",
                "sentinel-3-olci-wfr-l2-netcdf",
                "sentinel-3-synergy-v10-l2-netcdf",
                "sentinel-3-olci-lfr-l2-netcdf",
                "sentinel-3-slstr-lst-l2-netcdf",
                "sentinel-3-slstr-wst-l2-netcdf",
                "sentinel-3-synergy-syn-l2-netcdf",
                "sentinel-3-synergy-vgp-l2-netcdf",
                "sentinel-3-synergy-vg1-l2-netcdf",
            ],
            "safe-manifest",
        ),
        **dict.fromkeys(
            [
                "esa-cci-lc-netcdf",
                "noaa-climate-normals-netcdf",
                "noaa-cdr-sea-surface-temperature-whoi-netcdf",
                "noaa-cdr-ocean-heat-content-netcdf",
            ],
            "netcdf",
        ),
        **dict.fromkeys(
            [
                "noaa-mrms-qpe-24h-pass2",
                "noaa-mrms-qpe-1h-pass1",
                "noaa-mrms-qpe-1h-pass2",
            ],
            "cog",
        ),
        **dict.fromkeys(["landsat-c2-l2", "landsat-c2-l1"], "mtl.txt"),
        **dict.fromkeys(["sentinel-2-l2a"], "product-metadata"),
        **dict.fromkeys(["mtbs"], "burn-severity"),
        **dict.fromkeys(["alos-fnf-mosaic"], "C"),
        **dict.fromkeys(["nrcan-landcover"], "landcover"),
        **dict.fromkeys(["nasadem"], "elevation"),
        **dict.fromkeys(["naip"], "image"),
    }

    earth_search_map = {
        **dict.fromkeys(["sentinel-s2-l2a-cogs", "sentinel-2-l2a"], 1),
        **dict.fromkeys(
            ["sentinel-s2-l2a", "sentinel-s2-l1c", "sentinel-2-l1c"],
            ("visual", "productInfo.json"),
        ),
        **dict.fromkeys(["naip"], "image"),
        **dict.fromkeys(["landsat-c2-l2"], "mtl.txt"),
        **dict.fromkeys(["sentinel-1-grd"], "safe-manifest"),
        **dict.fromkeys(["cop-dem-glo-30", "cop-dem-glo-90"], "data"),
    }

    sentinel_hub_map = {
        **dict.fromkeys(["sentinel-2"], ("data", "productInfo.json")),
        **dict.fromkeys(["sentinel-1"], ("s3", "manifest.safe")),
    }

    product_file_map = {
        "planetarycomputer.microsoft.com/api/stac": planetary_computer_map,
        "earth-search.aws.element84.com": earth_search_map,
        "services.sentinel-hub.com/api": sentinel_hub_map,
    }

    stacs = list(product_file_map.keys())

    self_link = next(
        (link["href"] for link in item["links"] if link["rel"] == "self"), None
    )

    if self_link is None:
        return

    item_stac = next((stac for stac in stacs if stac in self_link), None)

    if item_stac is None:
        return

    collection_id = (
        item["collection"]
        if "collection" in item
        else (
            item["properties"]["constellation"] if item_stac == stacs[2] else item["id"]
        )
    )

    target = product_file_map[item_stac].get(collection_id)

    href = None
    if isinstance(target, str):
        href = (
            [
                cog["href"]
                for cog in item["assets"].values()
                if cog["href"].endswith((".tif", ".tiff"))
            ]
            if target == "All COGs"
            else item["assets"][target]["href"]
        )
    elif isinstance(target, int):
        href = item["links"][target]["href"]
    elif isinstance(target, tuple):
        directory = os.path.dirname(item["assets"][target[0]]["href"])
        if collection_id == "sentinel-s2-l2a":
            directory = os.path.dirname(directory)
        href = f"{directory}/{target[1]}"

    href = (
        rf"/vsis3{href[4:]}"
        if href is not None and isinstance(href, str) and href.startswith("s3")
        else href
    )

    if collection_id.startswith(("sentinel-2", "sentinel-s2", "landsat")):
        href = rf"{href}\Multiband"

    return href


def _get_stac_links(stac_json, rel):
    """
    This method is used to retrieve all the links matching the specified relation type from a STAC Item or Catalog.
    :param stac_json: input STAC Item or Catalog (JSON dictionary).
    :param rel: relationship type used to filter  the links.
    :return list (of URLs matching the rel filter)
    """
    if "links" not in stac_json:
        raise RuntimeError(f"Invalid STAC Item/Catalog-\n{stac_json}")
    links = stac_json["links"]
    rel_links = [l for l in links if l["rel"] == rel]
    link_hrefs = [l["href"] for l in rel_links]

    all_links = []
    for l in link_hrefs:
        if l.startswith("http"):
            link = l
        else:
            link = urljoin(os.path.dirname(stac_json["links"][0]["href"]) + "/", l)
        all_links.append(link)
    return all_links


def _get_all_stac_catalog_items(stac_json, request_params={}):
    """
    This method is used to get all items from a STAC catalog and all its subcatalogs. Will traverse any subcatalogs recursively.
    :param stac_json: input Static STAC (Catalog - JSON dictionary)
    :param request_params: requests.get() method parameters used for the STAC Item and Catalog requests (passed through the RasterCollection.from_stac_catalog() method call).
    :return generator (of all items retrived in the Catalog)
    """
    for item_link in _get_stac_links(stac_json, "item"):
        item_res = _requests.get(item_link, **request_params)
        if item_res.status_code != 200 or item_res.headers.get("content-type") not in [
            "application/json",
            "application/geo+json",
            "application/json;charset=utf-8",
        ]:
            raise RuntimeError(f"Invalid STAC Item-\n{item_res.text}")
        item_json = item_res.json()
        yield item_json

    children = _get_stac_links(stac_json, "child")
    for child in children:
        child_res = _requests.get(child, **request_params)
        if child_res.status_code != 200 or child_res.headers.get(
            "content-type"
        ) not in [
            "application/json",
            "application/geo+json",
            "application/json;charset=utf-8",
        ]:
            raise RuntimeError(f"Invalid STAC Catalog-\n{child_res.text}")
        child_json = child_res.json()
        yield from _get_all_stac_catalog_items(child_json, request_params)



import torch
import torch.nn as nn
import torch.nn.functional as F
from fastai.torch_core import Module
from functools import partial
from . import *

device = "cuda" if torch.cuda.is_available() else "cpu"


def noop(x):
    return x


class LambdaPlus(Module):
    def __init__(self, func, *args, **kwargs):
        self.func, self.args, self.kwargs = func, args, kwargs

    def forward(self, x):
        return self.func(x, *self.args, **self.kwargs)


def get_act_layer(act_fn, **kwargs):
    act_fn = act_fn.lower()
    assert act_fn in ["relu", "leakyrelu", "prelu", "elu"], "incorrect act_fn"
    if act_fn == "relu":
        return nn.ReLU()
    elif act_fn == "leakyrelu":
        return nn.LeakyReLU(**kwargs)
    elif act_fn == "prelu":
        return nn.PReLU(**kwargs)
    elif act_fn == "elu":
        return nn.ELU(**kwargs)


def same_padding1d(seq_len, ks, stride=1, dilation=1):
    assert stride > 0
    assert dilation >= 1
    effective_ks = (ks - 1) * dilation + 1
    out_dim = (seq_len + stride - 1) // stride
    p = max(0, (out_dim - 1) * stride + effective_ks - seq_len)
    padding_before = p // 2
    padding_after = p - padding_before
    return padding_before, padding_after


class ZeroPad1d(nn.ConstantPad1d):
    def __init__(self, padding):
        super().__init__(padding, 0.0)


class ConvSP1d(nn.Module):
    "Conv1d padding='same'"

    def __init__(
        self, c_in, c_out, ks, stride=1, padding="same", dilation=1, bias=True
    ):
        super().__init__()
        self.ks, self.stride, self.dilation = ks, stride, dilation
        self.conv = nn.Conv1d(
            c_in, c_out, ks, stride=stride, padding=0, dilation=dilation, bias=bias
        )
        self.zeropad = ZeroPad1d
        self.weight = self.conv.weight
        self.bias = self.conv.bias

    def forward(self, x):
        padding = same_padding1d(
            x.shape[-1], self.ks, stride=self.stride, dilation=self.dilation
        )
        return self.conv(self.zeropad(padding)(x))


def convlayer(
    c_in,
    c_out,
    ks=3,
    padding="same",
    bias=True,
    stride=1,
    bn_init=False,
    zero_bn=False,
    bn_before=True,
    act_fn="relu",
    **kwargs,
):
    """conv layer (padding="same") + bn + act"""
    if ks % 2 == 1 and padding == "same":
        padding = ks // 2
    layers = [
        ConvSP1d(c_in, c_out, ks, bias=bias, stride=stride)
        if padding == "same"
        else nn.Conv1d(c_in, c_out, ks, stride=stride, padding=padding, bias=bias)
    ]
    bn = nn.BatchNorm1d(c_out)
    if bn_init:
        nn.init.constant_(bn.weight, 0.0 if zero_bn else 1.0)
    if bn_before:
        layers.append(bn)
    if act_fn:
        layers.append(get_act_layer(act_fn, **kwargs))
    if not bn_before:
        layers.append(bn)
    return nn.Sequential(*layers)


class Flatten(nn.Module):
    def forward(self, x):
        return x.view(x.size(0), -1)


class Squeeze(nn.Module):
    def __init__(self, dim=-1):
        super().__init__()
        self.dim = dim

    def forward(self, x):
        return x.squeeze(dim=self.dim)


class Unsqueeze(nn.Module):
    def __init__(self, dim=-1):
        super().__init__()
        self.dim = dim

    def forward(self, x):
        return x.unsqueeze(dim=self.dim)


class YRange(nn.Module):
    def __init__(self, y_range: tuple):
        super().__init__()
        self.y_range = y_range

    def forward(self, x):
        x = F.sigmoid(x)
        x = x * (self.y_range[1] - self.y_range[0])
        return x + self.y_range[0]


class Mult(nn.Module):
    def __init__(self, mult, trainable=True):
        self.mult, self.trainable = mult, trainable
        super().__init__()
        self.weight = nn.Parameter(
            torch.Tensor(1).fill_(self.mult).to(device), requires_grad=self.trainable
        )

    def forward(self, x):
        return x.mul_(self.weight)


class Exp(nn.Module):
    def __init__(self, exp, trainable=True):
        self.exp, self.trainable = exp, trainable
        super().__init__()
        self.weight = nn.Parameter(
            torch.Tensor(1).fill_(self.exp).to(device), requires_grad=self.trainable
        )

    def forward(self, x):
        return x**self.weight


class GeneralRelu(nn.Module):
    def __init__(self, leak=None, sub=0.0, maxv=None, **kwargs):
        super().__init__()
        self.leak, self.sub, self.maxv = leak, sub, maxv

    def forward(self, x):
        x = F.leaky_relu(x, self.leak) if self.leak is not None else F.relu(x)
        x = x.sub_(self.sub)
        if self.maxv is not None:
            x.clamp_max_(self.maxv)
        return x


def get_act_fn_norm(act_fn):
    x = torch.randn(1000000)
    x = act_fn(x)
    x1 = x / x.std()
    x2 = x1 - x1.mean()
    return 1 / x.std(), x1.mean()


class AFN(nn.Module):
    def __init__(self, act_fn=F.relu, trainable=True, **kwargs):
        super().__init__()
        self.act_fn = partial(act_fn, **kwargs)
        mul, add = get_act_fn_norm(self.act_fn)
        self.weight = nn.Parameter(
            torch.Tensor(1).fill_(mul).to(device), requires_grad=trainable
        )
        self.bias = nn.Parameter(
            torch.Tensor(1).fill_(add).to(device), requires_grad=trainable
        )

    def forward(self, x):
        x = self.act_fn(x)
        x.mul_(self.weight)
        x.add_(self.bias)
        return x


class Permute(Module):
    def __init__(self, *dims):
        self.dims = dims

    def forward(self, x):
        return x.permute(self.dims)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(dims={', '.join([str(d) for d in self.dims])})"
        )


class Max(Module):
    def __init__(self, dim=None, keepdim=False):
        self.dim, self.keepdim = dim, keepdim

    def forward(self, x):
        return x.max(self.dim, keepdim=self.keepdim)[0]

    def __repr__(self):
        return f"{self.__class__.__name__}(dim={self.dim}, keepdim={self.keepdim})"


class Transpose(Module):
    def __init__(self, *dims, contiguous=False):
        self.dims, self.contiguous = dims, contiguous

    def forward(self, x):
        if self.contiguous:
            return x.transpose(*self.dims).contiguous()
        else:
            return x.transpose(*self.dims)

    def __repr__(self):
        if self.contiguous:
            return f"{self.__class__.__name__}(dims={', '.join([str(d) for d in self.dims])}).contiguous()"
        else:
            return (
                f"{self.__class__.__name__}({', '.join([str(d) for d in self.dims])})"
            )



import arcgis
from pathlib import Path
import os
import glob
import shutil
import time
import tempfile
import json
import logging
from .._data import _raise_fastai_import_error
from .._utils.env import (
    HAS_TENSORFLOW,
    raise_tensorflow_import_error,
    _LAMBDA_TEXT_CLASSIFICATION,
    is_arcgispronotebook,
    reload_IPython,
)
from warnings import warn
import contextlib
import io
import sys
import socket
from functools import wraps
import traceback
import inspect
import types
import functools
import warnings

HAS_FASTAI = True
HAS_TENSORBOARDX = True

try:
    if not _LAMBDA_TEXT_CLASSIFICATION:
        from fastai.vision.learner import model_meta, _default_meta
        from .._utils.common import get_post_processed_model
        from torchvision import models

    from fastai.callbacks import TrackerCallback, EarlyStoppingCallback
    from fastai.basic_train import LearnerCallback
    from torch import nn
    import torch
    import numpy as np
    import math
    import warnings
    from fastai.distributed import *
    from fastai.torch_core import distrib_barrier
    import argparse
    from torch.nn.parallel import DistributedDataParallel
    from .._utils.segmentation_loss_functions import dice
    from fastai.basics import partial
    import pandas as pd
    from ... import __version__ as ArcGISLearnVersion
    from ._pointcnn_utils import AverageMetric
    from fastai.core import camel2snake
    import timm
    from .._utils.evaluate_batchsize import estimate_batch_size
    from .._utils.evaluate_batchsize import unsupported_models
    from .._data import prepare_data

    # EarlyStoppingCallback should run as one
    # of the first callback so that stop training flag is set
    # and other callbacks can behave accordingly.
    # e.g: Do not checkpoint final model after early stopping.
    EarlyStoppingCallback._order = -10

except ImportError as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False

    class TrackerCallback:
        pass

    class LearnerCallback:
        pass


logger = logging.getLogger()

# For lr computation, skip beginning and trailing values.
losses_skipped = 5
trailing_losses_skipped = 5
model_characteristics_folder = "ModelCharacteristics"

if HAS_FASTAI and not _LAMBDA_TEXT_CLASSIFICATION:
    # Declare the family of backbones to be unpacked and used by different models as supported types
    _vgg_family = [
        models.vgg11.__name__,
        models.vgg11_bn.__name__,
        models.vgg13.__name__,
        models.vgg13_bn.__name__,
        models.vgg16.__name__,
        models.vgg16_bn.__name__,
        models.vgg19.__name__,
        models.vgg19_bn.__name__,
    ]
    _resnet_family = [
        models.resnet18.__name__,
        models.resnet34.__name__,
        models.resnet50.__name__,
        models.resnet101.__name__,
        models.resnet152.__name__,
    ]
    _densenet_family = [
        models.densenet121.__name__,
        models.densenet169.__name__,
        models.densenet161.__name__,
        models.densenet201.__name__,
    ]


@contextlib.contextmanager
def nostdout():
    save_stdout = sys.stdout
    sys.stdout = io.StringIO()
    yield
    sys.stdout = save_stdout


def _get_device():
    if getattr(arcgis.env, "_processorType", "") == "GPU" and torch.cuda.is_available():
        device = torch.device("cuda")
    elif getattr(arcgis.env, "_processorType", "") == "CPU":
        device = torch.device("cpu")
    else:
        device = (
            torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
        )

    return device


class _EmptyDS(object):
    def __init__(self, size):
        self.size = (size, size)


class _EmptyData:
    def __init__(self, path, c, loss_func, chip_size, train_ds=True):
        self.path = path.parent
        if (
            getattr(arcgis.env, "_processorType", "") == "GPU"
            and torch.cuda.is_available()
        ):
            self.device = torch.device("cuda")
        elif getattr(arcgis.env, "_processorType", "") == "CPU":
            self.device = torch.device("cpu")
        else:
            self.device = (
                torch.device("cuda")
                if torch.cuda.is_available()
                else torch.device("cpu")
            )
        self.c = c
        self.loss_func = loss_func
        self.chip_size = chip_size

        if train_ds:
            self.train_ds = [[_EmptyDS(chip_size)]]


class _MultiGPUCallback(LearnerCallback):
    """
    Parallelize over multiple GPUs only if multiple GPUs are present.
    """

    def __init__(self, learn):
        super(_MultiGPUCallback, self).__init__(learn)

        self.multi_gpu = torch.cuda.device_count() > 1

    def on_train_begin(self, **kwargs):
        if self.multi_gpu:
            logger.info("Training on multiple GPUs")
            self.learn.model = nn.DataParallel(self.learn.model)

    def on_train_end(self, **kwargs):
        if self.multi_gpu:
            self.learn.model = self.learn.model.module


def _set_multigpu_callback(model):
    if (
        (not hasattr(arcgis.env, "_gpuid"))
        or (arcgis.env._gpuid >= torch.cuda.device_count())
    ) and (not getattr(arcgis.env, "_processorType", False) == "CPU"):
        model.learn.callback_fns.append(_MultiGPUCallback)


def _set_ddp_multigpu(model):
    try:
        parser = argparse.ArgumentParser()
    except IndexError:
        model._multigpu_training = False
        return
    parser.add_argument("--local_rank", type=int)
    args, unknown = parser.parse_known_args()
    if "RANK" in os.environ and "WORLD_SIZE" in os.environ:
        args.rank = int(os.environ["RANK"])
        args.world_size = int(os.environ["WORLD_SIZE"])
        args.gpu = int(os.environ["LOCAL_RANK"])
    elif "SLURM_PROCID" in os.environ:
        args.rank = int(os.environ["SLURM_PROCID"])
        args.gpu = args.rank % torch.cuda.device_count()
    elif hasattr(args, "rank"):
        pass
    else:
        model._multigpu_training = False
        return
    model._multigpu_training = True
    args.gpu = args.gpu % torch.cuda.device_count()
    torch.cuda.set_device(args.gpu)
    backend = "nccl"
    if os.name == "nt":
        backend = "gloo"
    torch.distributed.init_process_group(
        backend=backend,
        init_method="env://",
        world_size=args.world_size,
        rank=args.rank,
    )
    torch.distributed.barrier()
    model._rank_distributed = args.gpu


def _isnotebook():
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True
        elif shell == "TerminalInteractiveShell":
            return False
        else:
            return False
    except NameError:
        return False


def _create_zip(zipname, path):
    import shutil

    if os.path.exists(os.path.join(path, zipname) + ".dlpk"):
        os.remove(os.path.join(path, zipname) + ".dlpk")

    temp_dir = tempfile.TemporaryDirectory().name
    zip_file = shutil.make_archive(os.path.join(temp_dir, zipname), "zip", path)
    dlpk_base = os.path.splitext(zip_file)[0]
    os.rename(zip_file, dlpk_base + ".dlpk")
    dlpk_file = dlpk_base + ".dlpk"
    shutil.move(dlpk_file, path)


class SaveModelCallback(TrackerCallback):
    def __init__(
        self,
        model,
        every="improvement",
        name="bestmodel",
        load_best_at_end=True,
        **kwargs,
    ):
        super().__init__(learn=model.learn, **kwargs)
        self.model = model
        self.every = every
        self.name = name
        self.load_best_at_end = load_best_at_end

        # set some default value of best epoch attribute
        self.best_epoch = 0
        self.learn._best_epoch = 0

        if self.every not in ["improvement", "epoch"]:
            warn(
                'SaveModel every {} is invalid, falling back to "improvement".'.format(
                    self.every
                )
            )
            self.every = "improvement"

    def on_epoch_end(self, epoch, **kwargs):
        "Compare the value monitored to its best score and maybe save the model."

        if int(os.environ.get("RANK", 0)):
            return

        # do not save model after early stopping kicks in.
        if not kwargs.get("stop_training", False):
            current = self.get_monitor_value()

            if isinstance(current, torch.Tensor):
                if current.is_cuda:
                    current = current.cpu()

            # if a better checkpoint is found.
            better_checkpoint = current is not None and self.operator(
                current, self.best
            )
            if better_checkpoint:
                self.best_epoch = epoch
                self.learn._best_epoch = epoch
                self.best = current

            self.current = current

            if self.every == "epoch":
                self.model._save(
                    f"{self.name}_epoch_{epoch}",
                    zip_files=False,
                    save_html=False,
                    compute_metrics=False,
                )
            # every improvement
            elif better_checkpoint:
                self.remove_previous()
                self.model._save(
                    f"{self.name}_epoch_{epoch}",
                    zip_files=False,
                    save_html=False,
                    compute_metrics=False,
                )

    def remove_previous(self):
        # to avoid creating multiple best checkpoints.
        saved_path = os.path.join(self.model.learn.path, self.model.learn.model_dir)
        for p in glob.glob(os.path.join(saved_path, self.name + "*")):
            shutil.rmtree(p)

    def on_train_end(self, **kwargs):
        "Load the best model."
        if self.load_best_at_end:
            try:
                self.model.load(f"{self.name}_epoch_{self.best_epoch}")
            except FileNotFoundError:
                # don't show message in child process in case of multigpu
                if not int(os.environ.get("RANK", 0)):
                    # logging this to notify about possible errors.
                    print("Could not load the best model.")

            try:
                self.model.save(
                    f"{self.name}_epoch_{self.best_epoch}", compute_metrics=False
                )
            except:
                # logging this to notify about possible errors.
                print("Encountered error in saving checkpoint.")


# Multispectral Models Specific resources start #

valid_init_schemes = ["red_band", "random", "all_random"]
rgb_map = {"r": 0, "red": 0, "g": 1, "green": 1, "b": 2, "blue": 2}


def get_band_mapping(band_name):
    # Extra Logic goes Here
    # For example: NIR --> RED(0); Coastal --> BLUE(2)
    # We can store Custom weights for multispectral models and load them in this logic
    #
    return rgb_map.get(band_name.lower(), None)


def _get_tail(model):
    if hasattr(model, "named_children"):
        child_name, child = next(model.named_children())
        if isinstance(child, nn.Conv2d):
            return child_name, child

    if hasattr(model, "children"):
        for children in model.children():
            try:
                child_name, child = _get_tail(children)
                return child_name, child
            except:
                pass


def _get_ms_tail(tail, data, type_init="random"):
    in_chanls = len(data._extract_bands)
    new_tail = nn.Conv2d(
        in_channels=in_chanls,
        out_channels=tail.out_channels,
        kernel_size=tail.kernel_size,
        stride=tail.stride,
        padding=tail.padding,
        dilation=tail.dilation,
        groups=tail.groups,
        bias=tail.bias is not None,
        padding_mode=tail.padding_mode,
    )
    # referred from https://github.com/rwightman/pytorch-image-models/blob/7c67d6aca992f039eece0af5f7c29a43d48c00e4/timm/models/helpers.py#L143
    if in_chanls == 1:
        new_tail.weight.data = tail.weight.data.float().sum(dim=1, keepdim=True)
    else:
        repeat = int(math.ceil(in_chanls / 3))
        new_tail.weight.data = (
            (tail.weight.data.float().repeat(1, repeat, 1, 1)[:, :in_chanls, :, :])
            * 3
            / float(in_chanls)
        )
    for i, j in enumerate(data._extract_bands):
        band = str(data._bands[j]).lower()
        b = get_band_mapping(band)  # rgb_map.get(band, None)
        if b is not None and not type_init == "all_random":
            new_tail.weight.data[:, i] = tail.weight.data[:, b]
        else:
            if type_init == "red_band":
                new_tail.weight.data[:, i] = tail.weight.data[
                    :, 0
                ]  # Red Band Weights for all other band weights
            elif type_init == "random" or type_init == "all_random":
                # Random Weights for all other band weights
                pass
    return new_tail


def _set_tail(model, new_tail):
    updated = False
    if hasattr(model, "named_children"):
        child_name, child = next(model.named_children())
        if isinstance(child, nn.Conv2d):
            setattr(model, child_name, new_tail)
            updated = True
    if hasattr(model, "children") and not updated:
        for children in model.children():
            try:
                _set_tail(children, new_tail)
                return
            except:
                pass


def _change_tail(model, data, tail_weights_type=None):
    tail_name, tail = _get_tail(model)
    if tail_weights_type is None:
        tail_weights_type = getattr(arcgis.env, "type_init_tail_parameters", "random")
    if tail_weights_type not in valid_init_schemes:
        raise Exception(
            f"""
        \n'{type_init}' is not a valid scheme for initializing model tail weights.
        \nplease set a valid scheme from 'red_band', 'random' or 'all_random'.
        \n`arcgis.env.type_init_tail_parameters={{valid_scheme}}`
        """
        )
    new_tail = _get_ms_tail(tail, data, type_init=tail_weights_type)
    _set_tail(model, new_tail)
    return model


def _get_backbone_meta(arch_name):
    _model_meta = {i.__name__: j for i, j in model_meta.items()}
    return _model_meta.get(arch_name, _default_meta)


# Multispectral Models Specific resources end #


def _device_check():
    if hasattr(arcgis, "env") and getattr(arcgis.env, "_processorType", "") == "CPU":
        return True

    if not torch.cuda.is_available():
        warnings.warn("Cuda is not available")
        return True

    move_to_cpu = False

    incorrect_binary_warn = """
    Found GPU%d %s which requires CUDA_VERSION >= %d for
    optimal performance and fast startup time, but your PyTorch was compiled
    with CUDA_VERSION %d. Please install the correct PyTorch binary
    using instructions from https://pytorch.org
    It may continue to work on CPU.
    """

    old_gpu_warn = """
    Found GPU%d %s which is of cuda capability %d.%d.
    PyTorch no longer supports this GPU because it is too old.
    The minimum cuda capability that we support is 3.7.
    Go to https://pytorch.org for more info on how to install
    or build a PyTorch version that has been compiled for your
    GPU architecture (Cuda compute capability).
    It may continue to work on CPU.
    """

    CUDA_VERSION = torch._C._cuda_getCompiledVersion()
    for d in range(torch.cuda.device_count()):
        capability = torch.cuda.get_device_capability(d)
        major = capability[0]
        name = torch.cuda.get_device_name(d)
        if CUDA_VERSION < 8000 and major >= 6:
            warnings.warn(incorrect_binary_warn % (d, name, 8000, CUDA_VERSION))
            move_to_cpu = True
        elif CUDA_VERSION < 9000 and major >= 7:
            warnings.warn(incorrect_binary_warn % (d, name, 9000, CUDA_VERSION))
            move_to_cpu = True
        elif capability == (3, 0) or major < 3:
            warnings.warn(old_gpu_warn % (d, name, major, capability[1]))
            move_to_cpu = True

    try:
        if not torch._C._cuda_isDriverSufficient():
            move_to_cpu = True
            if torch._C._cuda_getDriverVersion() == 0:
                # found no NVIDIA driver on the system
                warnings.warn(
                    """
                Found no GPU driver. CPU will be used for processing.
                """
                )
            else:
                warnings.warn(
                    """
                The NVIDIA driver on your system is too old (found version {})
                or your GPU architecture is very old and it's not supported.
                Please update your GPU driver by downloading and installing
                a new version from the URL: http://www.nvidia.com/Download/index.aspx
                Alternatively, go to https://pytorch.org for more info
                on how to install or build a PyTorch version that has been
                compiled for your GPU architecture (Cuda compute capability)
                or for your version of Cuda driver. It may continue to work on CPU.
                """.format(
                        str(torch._C._cuda_getDriverVersion())
                    )
                )
    except:
        pass

    return move_to_cpu


class ArcGISModel(object):
    def __init__(self, data, backbone=None, **kwargs):
        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)
        if not _LAMBDA_TEXT_CLASSIFICATION:
            move_to_cpu = _device_check()
        else:
            move_to_cpu = True
        # Force move to CPU
        if move_to_cpu:
            arcgis.env._processorType = "CPU"

        self._device = _get_device()

        if backbone is None:
            self._backbone = models.resnet34
        elif type(backbone) is str:
            if hasattr(models, backbone):
                self._backbone = getattr(models, backbone)
            elif hasattr(models.detection, backbone):
                self._backbone = getattr(models.detection, backbone)
            elif "timm:" in backbone:
                bckbn = backbone.split(":")[1]
                if hasattr(timm.models, bckbn):
                    self._backbone = getattr(timm.models, bckbn)
        else:
            self._backbone = backbone

        if not hasattr(self, "_backbone"):
            self._backbone = models.resnet34

        if hasattr(data, "_is_multispectral"):  # multispectral support
            self._is_multispectral = getattr(data, "_is_multispectral")
        else:
            self._is_multispectral = False
        if self._is_multispectral:
            self._imagery_type = data._imagery_type
            self._bands = data._bands
            self._orig_backbone = self._backbone

            @wraps(self._orig_backbone)
            def backbone_wrapper(*args, **inkwargs):
                if "pretrained_backbone" in kwargs:
                    pretrained_backbone = kwargs["pretrained_backbone"]
                    assert type(pretrained_backbone) == bool
                    if len(args) > 0:
                        args = tuple([pretrained_backbone, *args[1:]])
                    else:
                        inkwargs["pretrained"] = pretrained_backbone
                return _change_tail(
                    self._orig_backbone(*args, **inkwargs),
                    data,
                    kwargs.get("tail_weights_type"),
                )

            backbone_wrapper._is_multispectral = True
            self._backbone = backbone_wrapper

        if not hasattr(data, "class_mapping") and hasattr(data, "classes"):
            data.class_mapping = {v: v for v in data.classes}

        if data is not None and getattr(data, "path", None) is None:
            data.path = Path(os.path.abspath("."))

        if getattr(self, "_is_edge_detection", False):
            if len(data.classes) > 2:
                raise Exception(
                    "Found multi-labels in the data, This is a binary segmentation model and hence please export the data with binary labels."
                    # noqa
                )

            data.class_mapping = {1: data.classes[1]}

        self.learn = None
        self._data = data
        self._learning_rate = None
        self._backend = getattr(self, "_backend", "pytorch")
        self._model_metrics_cache = None
        self._slice_lr = True
        self._pretrained_path = kwargs.get("pretrained_path", None)
        if hasattr(self._data, "arcgis_init_kwargs"):
            self._check_data_support_with_pretrained_path()
        self._model_kwargs = kwargs
        if self.__class__.__name__ not in unsupported_models:
            if not getattr(data, "_is_empty", False) and hasattr(
                data, "_estimate_batch"
            ):
                if data._estimate_batch:
                    try:
                        data._estimate_batch = False
                        batch_size = estimate_batch_size(self, mode="none")
                        self._data.train_dl.batch_size = (
                            batch_size.recommended_batchsize
                        )
                        self._data.valid_dl.batch_size = (
                            batch_size.recommended_batchsize
                        )
                    except Exception as e:
                        data._estimate_batch = True

    def _check_data_support_with_pretrained_path(self):
        if self._data is not None and self._pretrained_path is not None:
            with open(Path(self._pretrained_path).with_suffix(".emd")) as f:
                emd = json.load(f)
            if self._data.chip_size != emd["ImageHeight"]:
                import copy
                from .._data import prepare_data
                import logging

                logger = logging.getLogger()
                logger.warning(
                    f"""Setting the `chip_size` of input data ({self._data.chip_size}) to same as input model's ({emd["ImageHeight"]})."""
                )
                arcgis_init_kwargs = copy.deepcopy(self._data.arcgis_init_kwargs)
                arcgis_init_kwargs["chip_size"] = emd["ImageHeight"]
                arcgis_init_kwargs["resize_to"] = emd["resize_to"]
                self._data = prepare_data(**arcgis_init_kwargs)

    def _check_backbone_support(self, backbone):
        "Fetches the backbone name and returns True if it is in the list of supported backbones"
        backbone_name = backbone if type(backbone) is str else backbone.__name__
        if type(backbone) is not str and "timm" in backbone.__module__:
            backbone_name = "timm:" + backbone.__name__
        return False if backbone_name not in self.supported_backbones else True

    def _check_dataset_support(self, data):
        "Fetches the dataset name and returns True if it is in the list of supported dataset type"
        if hasattr(data, "dataset_type"):
            if getattr(data, "dataset_type") not in self.supported_datasets:
                raise Exception(
                    f"Enter only compatible datasets from {', '.join(self.supported_datasets)}"
                )

    def _arcgis_init_callback(self):
        if self._is_multispectral:
            if self._data._train_tail:
                params_iterator = self.learn.model.parameters()
                next(
                    params_iterator
                ).requires_grad = True  # make first conv weights learnable

                tail_name, first_layer = _get_tail(self.learn.model)

                if (
                    first_layer.bias is not None
                    or self.__class__.__name__ == "MaskRCNN"
                    or self.__class__.__name__ == "ModelExtension"
                ):
                    # make first conv bias weights learnable
                    # In case of maskrcnn make the batch norm trainable
                    next(params_iterator).requires_grad = True
                self.learn.create_opt(slice(3e-3))
        if hasattr(self, "_show_results_multispectral"):
            self.show_results = self._show_results_multispectral

    # function for checking if data exists for using class functions.
    def _check_requisites(self):
        if isinstance(self._data, _EmptyData) or getattr(
            self._data, "_is_empty", False
        ):
            raise Exception("Can't call this function without data.")

    # function for checking if tensorflow is installed otherwise raise error.
    def _check_tf(self):
        if not HAS_TENSORFLOW:
            raise_tensorflow_import_error()

    def _init_tensorflow(self, data, backbone):
        self._check_tf()

        from .._utils.common import get_color_array
        from .._utils.common_tf import (
            handle_backbone_parameter,
            get_input_shape,
            check_backbone_is_mobile_optimized,
        )

        # Get color Array
        color_array = get_color_array(data.color_mapping)
        if len(data.color_mapping) == (data.c - 1):
            # Add Background color
            color_array = np.concatenate(
                [np.array([[0.0, 0.0, 0.0, 0.0]]), color_array]
            )
        data._multispectral_color_array = color_array

        # Handle Backbone
        self._backbone = handle_backbone_parameter(backbone)

        self._backbone_mobile_optimized = check_backbone_is_mobile_optimized(
            self._backbone
        )

        # Initialize Backbone
        in_shape = get_input_shape(data.chip_size)
        self._backbone_initalized = self._backbone(
            input_shape=in_shape, include_top=False, weights="imagenet"
        )

        self._backbone_initalized.trainable = False
        self._device = torch.device("cpu")
        self._data = data

    def lr_find(self, allow_plot=True):
        """
        Runs the Learning Rate Finder. Helps in choosing the
        optimum learning rate for training the model.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        allow_plot              Optional boolean. Display the plot of losses
                                against the learning rates and mark the optimal
                                value of the learning rate on the plot.
                                The default value is 'True'.
        =====================   ===========================================
        """
        self._check_requisites()
        temp1 = self.learn.path
        metrics = None
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            try:
                metrics = self.learn.metrics
                self.learn.metrics = []
                # ddp training
                if getattr(self, "_multigpu_training", False):
                    self.learn.lr_find()
                    distrib_barrier()
                    # remove tmp.pth created during lr_find in parent process
                    if not int(os.environ.get("RANK", 0)):
                        os.remove(
                            Path(self.learn.path) / self.learn.model_dir / "tmp.pth"
                        )
                else:
                    with tempfile.TemporaryDirectory(
                        prefix="arcgisTemp_"
                    ) as _tempfolder:
                        self.learn.path = Path(_tempfolder)
                        self.learn.lr_find()
            except Exception as e:
                # if some error comes in lr_find
                raise e
            finally:
                self.learn.metrics = metrics
                # Revert
                self.learn.path = temp1

            #
            self.learn.path = temp1

            from IPython.display import clear_output

            clear_output()
            lr, index = self._find_lr()
            if allow_plot:
                self._show_lr_plot(index)

        return lr

    def _show_lr_plot(
        self,
        index,
        losses_skipped=losses_skipped,
        trailing_losses_skipped=trailing_losses_skipped,
    ):
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(1, 1)
        losses = self.learn.recorder.losses
        lrs = self.learn.recorder.lrs
        final_losses_skipped = 0
        if (
            len(self.learn.recorder.losses[losses_skipped:-trailing_losses_skipped])
            >= 5
        ):
            losses = self.learn.recorder.losses[losses_skipped:-trailing_losses_skipped]
            lrs = self.learn.recorder.lrs[losses_skipped:-trailing_losses_skipped]
            final_losses_skipped = losses_skipped
        ax.plot(lrs, losses)
        ax.set_ylabel("Loss")
        ax.set_xlabel("Learning Rate")
        ax.set_xscale("log")
        ax.xaxis.set_major_formatter(plt.FormatStrFormatter("%.0e"))
        ax.plot(
            self.learn.recorder.lrs[index],
            self.learn.recorder.losses[index],
            markersize=10,
            marker="o",
            color="red",
        )

        plt.show()

    def _find_lr(
        self,
        losses_skipped=losses_skipped,
        trailing_losses_skipped=trailing_losses_skipped,
        section_factor=3,
    ):
        losses = self.learn.recorder.losses
        lrs = self.learn.recorder.lrs
        final_losses_skipped = 0
        if (
            len(self.learn.recorder.losses[losses_skipped:-trailing_losses_skipped])
            >= 5
        ):
            losses = self.learn.recorder.losses[losses_skipped:-trailing_losses_skipped]
            lrs = self.learn.recorder.lrs[losses_skipped:-trailing_losses_skipped]
            final_losses_skipped = losses_skipped

        n = len(losses)

        max_start = 0
        max_end = 0

        lds = [1] * n

        for i in range(1, n):
            for j in range(0, i):
                if losses[i] < losses[j] and lds[i] < lds[j] + 1:
                    lds[i] = lds[j] + 1
                if lds[max_end] < lds[i]:
                    max_end = i
                    max_start = max_end - lds[max_end]

        sections = (max_end - max_start) / section_factor
        final_index = max_start + int(sections) + int(sections / 2)
        return lrs[final_index], final_losses_skipped + final_index

    @property
    def _model_metrics(self):
        raise NotImplementedError

    @property
    def available_metrics(self):
        """
        List of available metrics that are displayed in the training
        table. Set `monitor` value to be one of these while calling
        the `fit` method.
        """
        metrics = ["valid_loss"]
        for m in self.learn.metrics:
            if isinstance(m, AverageMetric) or isinstance(m, functools.partial):
                metrics.append(m.func.__name__)
            elif isinstance(m, types.FunctionType):
                metrics.append(m.__name__)
            else:
                metrics.append(camel2snake(m.__class__.__name__))

        return metrics

    def fit(
        self,
        epochs=10,
        lr=None,
        one_cycle=True,
        early_stopping=False,
        checkpoint=True,  # "all", "best", True, False ("best" and True are same.)
        tensorboard=False,
        monitor="valid_loss",  # whatever is passed here, earlystopping and checkpointing will use that.
        **kwargs,
    ):
        """
        Train the model for the specified number of epochs and using the
        specified learning rates

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        epochs                  Required integer. Number of cycles of training
                                on the data. Increase it if underfitting.
        ---------------------   -------------------------------------------
        lr                      Optional float or slice of floats. Learning rate
                                to be used for training the model. If ``lr=None``,
                                an optimal learning rate is automatically deduced
                                for training the model.
        ---------------------   -------------------------------------------
        one_cycle               Optional boolean. Parameter to select 1cycle
                                learning rate schedule. If set to `False` no
                                learning rate schedule is used.
        ---------------------   -------------------------------------------
        early_stopping          Optional boolean. Parameter to add early stopping.
                                If set to 'True' training will stop if parameter
                                `monitor` value stops improving for 5 epochs.
                                A minimum difference of 0.001 is required for
                                it to be considered an improvement.
        ---------------------   -------------------------------------------
        checkpoint              Optional boolean or string.
                                Parameter to save checkpoint during training.
                                If set to `True` the best model
                                based on `monitor` will be saved during
                                training. If set to 'all', all checkpoints
                                are saved. If set to False, checkpointing will
                                be off. Setting this parameter loads the best
                                model at the end of training.
        ---------------------   -------------------------------------------
        tensorboard             Optional boolean. Parameter to write the training log.
                                If set to 'True' the log will be saved at
                                `<dataset-path>/training_log` which can be visualized in
                                tensorboard. Required tensorboardx version=2.1

                                The default value is 'False'.

                                .. note::
                                    Not applicable for Text Models
        ---------------------   -------------------------------------------
        monitor                 Optional string. Parameter specifies
                                which metric to monitor while checkpointing
                                and early stopping. Defaults to 'valid_loss'. Value
                                should be one of the metric that is displayed in
                                the training table. Use `{model_name}.available_metrics`
                                to list the available metrics to set here.
        =====================   ===========================================
        """
        if os.environ.get("BLOCK_MODEL_TRAINING", 0) == "1":
            raise Exception(f"This model cannot be trained in ArcGIS Online Notebooks")

        if getattr(self, "_is_mm3d", False):
            self.learn.model.prediction = False

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            self._check_requisites()

            if lr is None:
                print("Finding optimum learning rate.")

                lr = self.lr_find(allow_plot=False)
                if self._slice_lr is True and len(self.learn.layer_groups) > 1:
                    lr = slice(lr / 10, lr)

            self._learning_rate = lr
            self._model_metrics_cache = None
            if (
                getattr(self._data, "_dataset_type", None) == "Classified_Tiles"
                and (
                    dice.__qualname__
                    not in [
                        metric.func.__qualname__
                        if hasattr(metric, "func")
                        else metric.__qualname__
                        for metric in self.learn.metrics
                    ]
                )
                and not getattr(self, "_is_edge_detection", False)
            ):
                self.learn.metrics.extend([dice])
            if arcgis.env.verbose:
                logger.info("Fitting the model.")

            if (
                not (type(self).__name__) == "EfficientDet"
                and getattr(self, "_backend", "pytorch") == "tensorflow"
            ):
                checkpoint = False

            callbacks = kwargs["callbacks"] if "callbacks" in kwargs.keys() else []
            kwargs.pop("callbacks", None)
            monitored_names = self.available_metrics
            if monitor not in monitored_names:
                raise Exception(f"`monitor` must be set to one from {monitored_names}")
            self.monitor = monitor
            if early_stopping:
                callbacks.append(
                    EarlyStoppingCallback(
                        learn=self.learn, monitor=monitor, min_delta=0.001, patience=5
                    )
                )
            self._is_checkpointed = checkpoint
            if checkpoint:
                from datetime import datetime

                now = datetime.now()
                if checkpoint != True and checkpoint != "all":
                    raise Exception(
                        "Checkpoint can only be set to a boolean, or 'all'."
                    )
                every = "improvement" if checkpoint is True else "epoch"
                save_callback_params = kwargs.get(
                    "save_callback_params", {"monitor": monitor, "every": every}
                )
                callbacks.append(
                    SaveModelCallback(
                        self,
                        name=now.strftime("checkpoint_%Y-%m-%d_%H-%M-%S"),
                        **save_callback_params,
                    )
                )
            kwargs.pop("save_callback_params", None)
            # If tensorboardx is installed write a log with name as timestamp
            if tensorboard:
                try:
                    import tensorboardX

                    # LearnerTensorboardWriter uses SummaryWriter from tensorboardX
                    from fastai.callbacks.tensorboard import LearnerTensorboardWriter
                    from .._utils.tensorboard_utils import ArcGISTBCallback
                except:
                    raise
                training_id = time.strftime("log_%Y-%m-%d_%H-%M-%S")
                log_path = Path(self._data.path) / "training_log"
                abs_path = os.path.abspath(log_path)
                training_id = type(self).__name__ + "_" + training_id
                callbacks.append(
                    partial(
                        ArcGISTBCallback,
                        base_dir=log_path,
                        name=training_id,
                        arcgis_model=self,
                    )(learn=self.learn)
                )
                hostname = socket.gethostname()
                print(
                    "Monitor training on Tensorboard using the following command: 'tensorboard --host={} --logdir=\"{}\"'".format(
                        hostname, abs_path
                    )
                )
            # Send out a warning if tensorboardX is not installed
            elif tensorboard:
                warn(
                    "Install tensorboardX 2.1 'pip install tensorboardx==2.1' to write training log"
                )

            self._fit_callbacks = callbacks
            if one_cycle:
                self.learn.fit_one_cycle(epochs, lr, callbacks=callbacks, **kwargs)
            else:
                self.learn.fit(epochs, lr, callbacks=callbacks, **kwargs)

    def unfreeze(self):
        """
        Unfreezes the earlier layers of the model for fine-tuning.
        """
        self.learn.unfreeze()

    def plot_losses(self):
        """
        Plot validation and training losses after fitting the model.
        """
        self._check_requisites()
        try:
            self.learn.recorder.plot_losses()
        except:
            raise Exception("You need to train your model to compute losses")

    def _create_emd_template(
        self, path, compute_metrics=True, save_inference_file=True
    ):
        _emd_template = {}

        # For old models - add lr, ModelName
        if isinstance(self._data, _EmptyData) or getattr(
            self._data, "_is_empty", False
        ):
            _emd_template = self._data.emd
            _emd_template["ModelFormat"] = "NCHW"
            if self._backend == "tensorflow" and self._framework == "tflite":
                _emd_template["ModelFormat"] = "NHWC"
            _emd_template["ModelFile"] = path.name
            if not _emd_template.get("ModelName"):
                _emd_template["ModelName"] = type(self).__name__

            if not _emd_template.get("LearningRate"):
                _emd_template["LearningRate"] = "0.0"
            if _emd_template["ModelName"] in ["MaskRCNN", "UnetClassifier", "CycleGAN"]:
                _emd_template["SupportsVariableTileSize"] = True
            else:
                _emd_template["SupportsVariableTileSize"] = False
            _emd_template["ArcGISLearnVersion"] = ArcGISLearnVersion

            return _emd_template

        if self._backbone is None:
            backbone = self._backbone
        else:
            if self._backend == "tensorflow":
                backbone = self._backbone._keras_api_names[-1].split(".")[-1]
            else:
                if "timm" in self._backbone.__module__:
                    backbone = "timm:" + self._backbone.__name__
                else:
                    backbone = self._backbone.__name__
            if backbone == "backbone_wrapper":
                backbone = self._orig_backbone.__name__

        _emd_template = self._get_emd_params(save_inference_file)

        _emd_template["ModelFormat"] = "NCHW"
        if self._backend == "tensorflow" and self._framework == "tflite":
            _emd_template["ModelFormat"] = "NHWC"
        if getattr(self, "_data", None) is not None:
            _emd_template["MinCellSize"] = getattr(self._data, "_emd", {}).get(
                "MinCellSize", None
            )
            _emd_template["MaxCellSize"] = getattr(self._data, "_emd", {}).get(
                "MaxCellSize", None
            )

        _emd_template["SupportsVariableTileSize"] = _emd_template.get(
            "SupportsVariableTileSize", False
        )
        _emd_template["ArcGISLearnVersion"] = ArcGISLearnVersion

        if getattr(self, "_fit_callbacks", None) is not None:
            checkpoint_callback = [
                c for c in self._fit_callbacks if isinstance(c, SaveModelCallback)
            ]
            if checkpoint_callback != []:
                checkpoint_callback = checkpoint_callback[0]
                key = getattr(self, "monitor", "valid_loss")
                val = checkpoint_callback.best
                if isinstance(val, torch.Tensor):
                    val = val.cpu().item()
                else:
                    val = float(val)
                _emd_template[f"monitored_{key}"] = val

        if isinstance(self._learning_rate, slice):
            _emd_lr = slice(
                "{0:1.4e}".format(self._learning_rate.start),
                "{0:1.4e}".format(self._learning_rate.stop),
            )
        elif self._learning_rate is not None:
            _emd_lr = "{0:1.4e}".format(self._learning_rate)
        else:
            _emd_lr = None

        _emd_template["ModelFile"] = path.name

        if hasattr(self._data, "chip_size"):
            chip_size = self._data.chip_size
            if not isinstance(chip_size, tuple):
                chip_size = (chip_size, chip_size)

            _emd_template["ImageHeight"] = chip_size[0]
            _emd_template["ImageWidth"] = chip_size[1]

        if hasattr(self._data, "_image_space_used"):
            _emd_template["ImageSpaceUsed"] = self._data._image_space_used

        _emd_template["LearningRate"] = str(_emd_lr)
        _emd_template["ModelName"] = type(self).__name__.replace("_", "")
        _emd_template["backend"] = self._backend

        if getattr(self, "_is_mmsegdet", False):
            model_params = {
                "model_name": self._kwargs["model"],
                "backend": self._backend,
            }
        elif getattr(self, "model_type", False) == "SR3":
            model_params = {"backbone": "SR3", "backend": self._backend}
        else:
            model_params = {"backbone": backbone, "backend": self._backend}
        if _emd_template.get("ModelParameters", None) is None:
            _emd_template["ModelParameters"] = model_params
        else:
            for _key in model_params:
                _emd_template["ModelParameters"][_key] = model_params[_key]

        if compute_metrics:
            if self._model_metrics_cache == None:
                print("Computing model metrics...")
                self._model_metrics_cache = self._model_metrics
            _emd_template.update(self._model_metrics_cache)

        resize_to = None
        if hasattr(self._data, "resize_to") and self._data.resize_to:
            resize_to = self._data.resize_to

        _emd_template["resize_to"] = resize_to

        # Check if model is Multispectral and dump parameters for that
        _emd_template["IsMultispectral"] = getattr(self, "_is_multispectral", False)
        if _emd_template.get("IsMultispectral", False):
            _emd_template["Bands"] = self._data._bands
            _emd_template["ImageryType"] = self._data._imagery_type
            if getattr(self._data, "_dataset_type", None) != "ChangeDetection":
                _emd_template["ExtractBands"] = self._data._extract_bands
            if not getattr(self._data, "_dataset_type", None) == "SuperResolution":
                _emd_template["NormalizationStats"] = {
                    "band_min_values": self._data._band_min_values,
                    "band_max_values": self._data._band_max_values,
                    "band_mean_values": self._data._band_mean_values,
                    "band_std_values": self._data._band_std_values,
                    "scaled_min_values": self._data._scaled_min_values,
                    "scaled_max_values": self._data._scaled_max_values,
                    "scaled_mean_values": self._data._scaled_mean_values,
                    "scaled_std_values": self._data._scaled_std_values,
                }
                for _stat in _emd_template["NormalizationStats"]:
                    if _emd_template["NormalizationStats"][_stat] is not None:
                        _emd_template["NormalizationStats"][_stat] = _emd_template[
                            "NormalizationStats"
                        ][_stat].tolist()
                _emd_template["DoNormalize"] = self._data._do_normalize
        if (
            getattr(self._data, "_dataset_type", None) == "Pix2Pix"
            or getattr(self._data, "_dataset_type", None) == "CycleGAN"
            or getattr(self._data, "_dataset_type", None) == "WNet_cGAN"
        ):
            _emd_template["ExtractBands"] = self._data._extract_bands
            _emd_template["NormalizationStats"] = {
                "band_min_values": self._data._band_min_values,
                "band_max_values": self._data._band_max_values,
                "band_mean_values": self._data._band_mean_values,
                "band_std_values": self._data._band_std_values,
                "scaled_min_values": self._data._scaled_min_values,
                "scaled_max_values": self._data._scaled_max_values,
                "scaled_mean_values": self._data._scaled_mean_values,
                "scaled_std_values": self._data._scaled_std_values,
            }
            for _stat in _emd_template["NormalizationStats"]:
                if _emd_template["NormalizationStats"][_stat] is not None:
                    _emd_template["NormalizationStats"][_stat] = _emd_template[
                        "NormalizationStats"
                    ][_stat].tolist()
            if getattr(self._data, "_dataset_type", None) == "WNet_cGAN":
                _emd_template["NormalizationStats_b"] = {
                    "band_min_values": self._data._band_min_values_b,
                    "band_max_values": self._data._band_max_values_b,
                    "band_mean_values": self._data._band_mean_values_b,
                    "band_std_values": self._data._band_std_values_b,
                    "scaled_min_values": self._data._scaled_min_values_b,
                    "scaled_max_values": self._data._scaled_max_values_b,
                    "scaled_mean_values": self._data._scaled_mean_values_b,
                    "scaled_std_values": self._data._scaled_std_values_b,
                }
                for _stat in _emd_template["NormalizationStats_b"]:
                    if _emd_template["NormalizationStats_b"][_stat] is not None:
                        _emd_template["NormalizationStats_b"][_stat] = _emd_template[
                            "NormalizationStats_b"
                        ][_stat].tolist()
            if getattr(self._data, "_dataset_type", None) == "CycleGAN":
                _emd_template["n_channel_rev"] = len(
                    _emd_template["NormalizationStats"]["band_min_values"]
                )
        if getattr(self._data, "_dataset_type", None) == "Classified_Tiles":
            if not getattr(self, "_is_edge_detection", False):
                if not getattr(self, "_orient_data", False):
                    if compute_metrics:
                        _emd_template[
                            "per_class_metrics"
                        ] = self.per_class_metrics().to_json()
        return _emd_template

    @staticmethod
    def _write_emd(_emd_template, path):
        json.dump(_emd_template, open(path, "w"), indent=4)

        return path.stem

    def _get_emd_params(self, save_inference_file):
        return {}

    @staticmethod
    def _create_html(path_model):
        import base64

        model_characteristics_dir = os.path.join(
            path_model.parent.absolute(), model_characteristics_folder
        )
        loss_graph = os.path.join(model_characteristics_dir, "loss_graph.png")
        show_results = os.path.join(model_characteristics_dir, "show_results.png")
        confusion_matrix = os.path.join(
            model_characteristics_dir, "confusion_matrix.png"
        )
        metrics_file = os.path.join(model_characteristics_dir, "metrics.html")
        results_file = os.path.join(model_characteristics_dir, "results.html")
        iframe_showresults = os.path.exists(
            os.path.join(model_characteristics_dir, "show_results.html")
        )

        encoded_losses_img = None
        if os.path.exists(loss_graph):
            encoded_losses_img = "data:image/png;base64,{0}".format(
                base64.b64encode(open(loss_graph, "rb").read()).decode("utf-8")
            )

        encoded_showresults = None
        if os.path.exists(show_results):
            encoded_showresults = "data:image/png;base64,{0}".format(
                base64.b64encode(open(show_results, "rb").read()).decode("utf-8")
            )

        confusion_matrix_img = None
        if os.path.exists(confusion_matrix):
            confusion_matrix_img = "data:image/png;base64,{0}".format(
                base64.b64encode(open(confusion_matrix, "rb").read()).decode("utf-8")
            )

        metrics_html = None
        if os.path.exists(metrics_file):
            with open(metrics_file, "r") as f:
                metrics_html = f.read()

        results_html = None
        if os.path.exists(results_file):
            with open(results_file, "r") as f:
                results_html = f.read()

        html_file_path = os.path.join(path_model.parent, "model_metrics.html")

        emd_path = os.path.join(path_model.parent, path_model.stem + ".emd")
        if not os.path.exists(emd_path):
            return

        emd_template = json.load(open(emd_path, "r"))

        encoded_losses = ""
        if encoded_losses_img:
            encoded_losses = f"""
                <p><b>Training and Validation loss</b></p>
                <img src="{encoded_losses_img}" alt="training and validation losses">
            """

        if emd_template.get("ModelParameters", {}).get("model_name", False):
            HTML_TEMPLATE = f"""        
                <p><b> {emd_template.get("ModelName").replace('>', '').replace('<', '')} </b></p>
                <p><b>Model Name:</b> {emd_template.get('ModelParameters', {}).get('model_name')}</p>
                <p><b>Learning Rate:</b> {emd_template.get('LearningRate')}</p>
                {encoded_losses}
            """

        else:
            HTML_TEMPLATE = f"""        
                    <p><b> {emd_template.get("ModelName").replace('>', '').replace('<', '')} </b></p>
                    <p><b>Backbone:</b> {emd_template.get('ModelParameters', {}).get('backbone')}</p>
                    <p><b>Learning Rate:</b> {emd_template.get('LearningRate')}</p>
                    {encoded_losses}
            """

        model_analysis = None
        if confusion_matrix_img:
            model_analysis = f"""
                    <img src="{confusion_matrix_img}" alt="Confusion Matrix">
            """

        if emd_template.get("accuracy"):
            model_analysis = f"""
            <p><b>Accuracy:</b> {emd_template.get('accuracy')}</p>
        """
        if emd_template.get("mIoU"):
            model_analysis = f"""
            <p><b>mIoU:</b> {emd_template.get('mIoU')}</p>
        """

        if emd_template.get("average_precision_score"):
            model_analysis = f"""
            <p><b>Average Precision Score:</b> {emd_template.get('average_precision_score')}</p>
        """

        if emd_template.get("score"):
            model_analysis = f"""
            <p><b>Score:</b> {emd_template.get('score')}</p>
        """

        if emd_template.get("PSNR"):
            model_analysis = f"""
            <p><b>PSNR Metric:</b> {emd_template.get('PSNR')}</p>
            <p><b>SSIM Metric:</b> {emd_template.get('SSIM')}</p>
        """
            # FID is supported for RGB only
            if emd_template.get("FID"):
                model_analysis = (
                    model_analysis
                    + f"""
                <p><b>FID Metric:</b> {emd_template.get('FID')}</p>
                """
                )

        if emd_template.get("per_class_metrics"):
            html_table = pd.read_json(emd_template.get("per_class_metrics")).to_html()
            model_analysis = f"""
            <p><b>Per class metrics:</b> {html_table}</p>
        """
        if emd_template.get("FID_A"):
            model_analysis = f"""
            <p><b>FID A:</b> {emd_template.get('FID_A')}</p>
            <p><b>FID B:</b> {emd_template.get('FID_B')}</p>
        """
        if emd_template.get("panoptic_quality"):
            model_analysis = f"""
            <p><b>Panoptic Quality:</b> {emd_template.get('panoptic_quality')}</p>
        """

        if model_analysis:
            HTML_TEMPLATE += f"""
            <p><b>Analysis of the model</b></p>
            {model_analysis}
        """

        if encoded_showresults:
            HTML_TEMPLATE += f"""
                <p><b>Sample Results</b></p>
                <img src="{encoded_showresults}" alt="Sample Results">
            """

        # For PointCNN and 3d models.
        if iframe_showresults:
            HTML_TEMPLATE += f"""
                <p><b>Sample Results</b><p>
                <iframe src="ModelCharacteristics/show_results.html" style="width:100%;height:70%;" scrolling="no" frameborder="0">
                    </iframe>
            """

        if metrics_html:
            HTML_TEMPLATE += f"""
                <p><b>Metrics per label</b></p>
                {metrics_html}
            """

        if results_html:
            HTML_TEMPLATE += f"""
                <p><b>Sample Results</b></p>
                {results_html}
            """

        file = open(html_file_path, "w")
        file.write(HTML_TEMPLATE)
        file.close()

    def _save(
        self,
        name_or_path,
        framework="PyTorch",
        zip_files=True,
        save_html=True,
        publish=False,
        gis=None,
        compute_metrics=True,
        save_optimizer=False,
        save_inference_file=True,
        **kwargs,
    ):
        save_format = kwargs.get("save_format", "default")  # 'default', 'tflite'
        post_processed = kwargs.get("post_processed", True)  # True, False
        quantized = kwargs.get("quantized", False)  # True, False
        temp = self.learn.path
        temp1 = self.learn.model_dir
        self._framework = framework
        if "\\" in name_or_path or "/" in name_or_path:
            path = Path(name_or_path)
            name = path.parts[-1]
            # to make fastai save to both path and with name
            self.learn.path = path
            self.learn.model_dir = ""
            if not os.path.exists(self.learn.path):
                os.makedirs(self.learn.path)
        else:
            # fixing fastai bug
            # self.learn.path = self.learn.path.parent
            self.learn.model_dir = Path(self.learn.model_dir) / name_or_path
            if not os.path.exists(self.learn.path / self.learn.model_dir):
                os.makedirs(self.learn.path / self.learn.model_dir)
            name = name_or_path

        script_paths = []
        onnx_paths = []
        tflite_paths = []
        try:
            _framework = framework.lower()
            if self._backend == "tensorflow" and _framework == "tflite":
                saved_path = self._save_tflite(
                    name, post_processed=post_processed, quantized=quantized
                )
            else:
                if self._backend != "tensorflow" and _framework == "tflite":
                    supported_models = [
                        "FeatureClassifier",
                    ]
                    if (type(self).__name__) in supported_models:
                        with warnings.catch_warnings():
                            warnings.simplefilter("ignore")
                            tflite_paths = self._save_pytorch_tflite(name)
                    else:
                        raise Exception(
                            "This pytorch model cannot be saved in tflite format"
                        )
                if self._backend == "pytorch" and _framework == "torchscript":
                    supported_models = [
                        "MaskRCNN",
                        "SingleShotDetector",
                        "YOLOv3",
                        "RetinaNet",
                        "SiamMask",
                    ]
                    if (type(self).__name__) in supported_models:
                        if type(self).__name__ != "SiamMask":
                            with warnings.catch_warnings():
                                warnings.simplefilter("ignore")
                                script_paths = self._save_pytorch_torchscript(name)
                    else:
                        raise Exception(
                            "This pytorch model cannot be saved in torchscript format"
                        )

                if isinstance(self.learn.model, (DistributedDataParallel)):
                    if not int(os.environ.get("RANK", 0)):
                        saved_path = self.learn.save(
                            name, return_path=True, with_opt=save_optimizer
                        )
                    return

                saved_path = self.learn.save(
                    name, return_path=True, with_opt=save_optimizer
                )

            # undoing changes to self.learn.path
        except Exception as e:
            raise e
        finally:
            self.learn.path = temp
            self.framework = framework
            self.learn.model_dir = temp1

        if (type(self).__name__) == "EfficientDet":
            _emd_template = self._create_emd_template(
                saved_path, compute_metrics, save_inference_file
            )
        else:
            _emd_template = self._create_emd_template(
                saved_path.with_suffix(".pth"), compute_metrics, save_inference_file
            )

        if framework.lower() == "tf-onnx":
            batch_size = kwargs.get("batch_size", 16)

            with nostdout():
                self._save_as_tfonnx(saved_path, batch_size)

            self._create_tfonnx_emd_template(
                _emd_template, saved_path.with_suffix(".onnx"), batch_size
            )
            os.remove(saved_path.with_suffix(".pth"))

        if self._backend != "tensorflow" and framework.lower() == "tflite":
            if len(tflite_paths) != 0:
                _script_save_params = {"tf": tflite_paths[0], "sm": tflite_paths[1]}
                _emd_template["TFLite"] = _script_save_params

        # TODO: merge all
        if framework.lower() == "torchscript":
            if len(script_paths) != 0:  # TODO: change_siammask
                _script_save_params = {"GPU": script_paths[1], "CPU": script_paths[0]}
                _emd_template["TorchScript"] = _script_save_params
            else:
                from ._siammask_utils import Custom
                from ._siammask_utils import load_pretrain

                siammask = Custom(anchors=self.anchors)
                if "\\" in name_or_path or "/" in name_or_path:
                    models_path = os.path.join(name_or_path)
                else:
                    models_path = os.path.join(
                        self.learn.path, self.learn.model_dir, name
                    )
                if not os.path.exists(models_path):
                    os.makedirs(models_path)

                siammask = load_pretrain(
                    siammask, os.path.join(models_path, name + ".pth")
                )
                outdir = os.path.join(models_path, "torch_scripts")
                if not os.path.isdir(outdir):
                    os.mkdir(outdir)

                scripted_feature_extractor = torch.jit.script(
                    siammask.features.features
                )
                scripted_feature_extractor.save(
                    os.path.join(outdir, "feature_extractor.pt")
                )

                scripted_feature_downsampler = torch.jit.script(
                    siammask.features.downsample
                )
                scripted_feature_downsampler.save(
                    os.path.join(outdir, "feature_downsampler.pt")
                )

                scripted_rpn_model = torch.jit.script(siammask.rpn_model)
                scripted_rpn_model.save(os.path.join(outdir, "rpn_model.pt"))

                scripted_mask_conv_kernel = torch.jit.script(
                    siammask.mask_model.mask.conv_kernel
                )
                scripted_mask_conv_kernel.save(
                    os.path.join(outdir, "mask_conv_kernel.pt")
                )

                scripted_mask_conv_search = torch.jit.script(
                    siammask.mask_model.mask.conv_search
                )
                scripted_mask_conv_search.save(
                    os.path.join(outdir, "mask_conv_search.pt")
                )

                scripted_mask_depthwise_conv = torch.jit.script(
                    siammask.mask_model.mask.conv2d_dw_group
                )
                scripted_mask_depthwise_conv.save(
                    os.path.join(outdir, "mask_depthwise_conv.pt")
                )

                scripted_refine_model = torch.jit.script(siammask.refine_model)
                scripted_refine_model.save(os.path.join(outdir, "refine_model.pt"))
                temp_emd_template = _emd_template.copy()
                temp_emd_template["ModelFile"] = "."
                temp_emd_template["ModelFiles"] = [
                    "feature_extractor.pt",
                    "feature_downsampler.pt",
                    "rpn_model.pt",
                    "mask_conv_kernel.pt",
                    "mask_conv_search.pt",
                    "mask_depthwise_conv.pt",
                    "refine_model.pt",
                ]

                if os.path.exists(os.path.join(outdir, name + ".emd")):
                    os.remove(os.path.join(outdir, name + ".emd"))

                import zipfile

                dlpk_Name = os.path.join(outdir, name + ".dlpk")
                if os.path.exists(dlpk_Name):
                    os.remove(dlpk_Name)

                out_file = open(os.path.join(outdir, name + ".emd"), "w")
                json.dump(temp_emd_template, out_file, indent=4)
                out_file.close()
                dlpk_Name = os.path.join(outdir, name + ".dlpk")
                f = zipfile.ZipFile(dlpk_Name, "w")
                cwd = os.getcwd()
                os.chdir(outdir)
                for files in temp_emd_template["ModelFiles"]:
                    f.write(files)

                f.write(name + ".emd")
                f.close()
                os.chdir(cwd)

        if _emd_template.get("InferenceFunction", False):
            if (
                _emd_template["ModelType"]
                not in [
                    "ObjectDetection",
                    "ImageClassification",
                    "InstanceDetection",
                    "ObjectClassification",
                    "CycleGAN",
                    "Pix2Pix",
                    "SuperResolution",
                    "ImageCaptioner",
                    "PanopticSegmenter",
                ]
                or save_inference_file
            ):
                inference_file = _emd_template["InferenceFunction"]
                if "[Functions]" in inference_file:
                    inference_file = inference_file[
                        len("[Functions]System\\DeepLearning\\ArcGISLearn\\") :
                    ]
                    _emd_template["InferenceFunction"] = inference_file

                with open(
                    saved_path.parent / _emd_template["InferenceFunction"], "w"
                ) as f:
                    f.write(self._code)
            if not save_inference_file:
                inference_file = _emd_template["InferenceFunction"]
                if "[Functions]" not in inference_file:
                    _emd_template["InferenceFunction"] = (
                        "[Functions]System\\DeepLearning\\ArcGISLearn\\"
                        + _emd_template["InferenceFunction"]
                    )

        ArcGISModel._write_emd(_emd_template, saved_path.with_suffix(".emd"))
        zip_name = saved_path.stem

        if save_html:
            # Backup env var
            bak_IS_ARCGISPRONOTEBOOK = arcgis.learn._utils.env._IS_ARCGISPRONOTEBOOK
            arcgis.learn._utils.env.switch = False
            try:
                # Do not call plt.show()
                arcgis.learn._utils.env._IS_ARCGISPRONOTEBOOK = False
                #
                self._save_model_characteristics(
                    saved_path.parent.absolute() / model_characteristics_folder
                )
                ArcGISModel._create_html(saved_path)
            except:
                pass
            finally:
                # Restore env var
                arcgis.learn._utils.env._IS_ARCGISPRONOTEBOOK = bak_IS_ARCGISPRONOTEBOOK
                is_arcgispronotebook()

        if _emd_template.get("ModelConfigurationFile", False):
            with open(
                saved_path.parent / _emd_template["ModelConfigurationFile"], "w"
            ) as f:
                f.write(inspect.getsource(self._model_conf_class))

        if zip_files:
            _create_zip(str(zip_name), str(saved_path.parent))

        if arcgis.env.verbose:
            print("Created model files at {spp}".format(spp=saved_path.parent))

        if publish:
            self._publish_dlpk(
                (saved_path.parent / os.path.basename(saved_path)).with_suffix(".dlpk"),
                gis=gis,
                overwrite=kwargs.get("overwrite", False),
            )

        return saved_path.parent

    def _save_tflite(self, name, post_processed=True, quantized=False):
        if post_processed or quantized:
            input_normalization = quantized is False
            return self.learn._save_tflite(
                name,
                return_path=True,
                model_to_save=self._get_post_processed_model(
                    input_normalization=input_normalization
                ),
                quantized=quantized,
                data=self._data,
            )
        return self.learn._save_tflite(name)

    def _save_pytorch_tflite(self, name):
        pass

    def _script(self, model, inp):
        scripted_model = torch.jit.script(model, inp)
        scripted_model.eval()
        return scripted_model

    def _trace(self, model, inp, check_trace=False):
        traced_model = torch.jit.trace(model, inp, check_trace=check_trace)
        traced_model.eval()
        return traced_model

    def _save_pytorch_torchscript(self, name):
        pass

    def _get_post_processed_model(self, input_normalization=True):
        return get_post_processed_model(self, input_normalization=input_normalization)

    def _save_model_characteristics(self, model_characteristics_dir):
        import shutil
        import matplotlib.pyplot as plt

        if isinstance(self._data, _EmptyData) or getattr(
            self._data, "_is_empty", False
        ):
            if not os.path.exists(
                os.path.join(self._data.emd_path.parent, model_characteristics_folder)
            ):
                return
            temp_path = tempfile.NamedTemporaryFile().name
            shutil.copytree(
                os.path.join(self._data.emd_path.parent, model_characteristics_folder),
                temp_path,
            )
            if os.path.exists(
                os.path.join(model_characteristics_dir, model_characteristics_dir)
            ):
                shutil.rmtree(
                    os.path.join(model_characteristics_dir, model_characteristics_dir),
                    ignore_errors=True,
                )

            shutil.copytree(
                temp_path,
                os.path.join(model_characteristics_dir, model_characteristics_dir),
            )

            return

        if not os.path.exists(
            os.path.join(model_characteristics_dir, model_characteristics_dir)
        ):
            os.mkdir(os.path.join(model_characteristics_dir, model_characteristics_dir))

        if hasattr(self.learn, "recorder"):
            try:
                self.learn.recorder.plot_losses()
                plt.savefig(os.path.join(model_characteristics_dir, "loss_graph.png"))
                plt.close()
            except:
                plt.close()

        if self.__str__() in [
            "<PointCNN>",
            "<RandLANet>",
            "<SQNSeg>",
            "<MMDetection3D>",
        ]:
            self.show_results(save_html=True, save_path=model_characteristics_dir)
        elif self.__str__() in [
            "<TextClassifier>",
            "<TransformerEntityRecognizer>",
            "<SequenceToSequence>",
            "<TimeSeriesModel>",
        ]:
            pass
        elif hasattr(self, "show_results"):
            self.show_results()
            plt.savefig(os.path.join(model_characteristics_dir, "show_results.png"))
            plt.close()

        if hasattr(self, "_save_confusion_matrix"):
            self._save_confusion_matrix(model_characteristics_dir)

    def _publish_dlpk(self, dlpk_path, gis=None, overwrite=False):
        gis_user = arcgis.env.active_gis if gis is None else gis
        if not gis_user:
            warn("No active gis user found!")
            return

        if not os.path.exists(dlpk_path):
            warn("DLPK file not found!")
            return

        emd_path = os.path.join(dlpk_path.parent, dlpk_path.stem + ".emd")

        if not os.path.exists(emd_path):
            warn("EMD File not found!")
            return

        emd_data = json.load(open(emd_path, "r"))
        formatted_description = f"""
                <p><b> {emd_data.get('ModelName').replace('>', '').replace('<', '')} </b></p>
                <p><b>Backbone:</b> {emd_data.get('ModelParameters', {}).get('backbone')}</p>
                <p><b>Learning Rate:</b> {emd_data.get('LearningRate')}</p>
        """

        if emd_data.get("accuracy"):
            formatted_description = (
                formatted_description
                + f"""
                <p><b>Analysis of the model</b></p>
                <p><b>Accuracy:</b> {emd_data.get('accuracy')}</p>
            """
            )

        if emd_data.get("average_precision_score"):
            formatted_description = (
                formatted_description
                + f"""
                <p><b>Analysis of the model</b></p>
                <p><b>Average Precision Score:</b> {emd_data.get('average_precision_score')}</p>
            """
            )

        item = gis_user.content.add(
            {
                "type": "Deep Learning Package",
                "description": formatted_description,
                "title": dlpk_path.stem,
                "overwrite": "true" if overwrite else "false",
            },
            data=str(dlpk_path.absolute()),
        )

        print(f"Published DLPK Item Id: {item.itemid}")

        model_characteristics_dir = os.path.join(
            dlpk_path.parent.absolute(), model_characteristics_folder
        )
        screenshots = [
            os.path.join(model_characteristics_dir, screenshot)
            for screenshot in os.listdir(model_characteristics_dir)
        ]

        item.update(item_properties={"screenshots": screenshots})

    def _create_tfonnx_emd_template(self, _emd_template, saved_path, batch_size):
        _emd_template.update(self._get_tfonnx_emd_params())
        _emd_template["BatchSize"] = batch_size
        _emd_template["ModelFile"] = saved_path.name

        return _emd_template

    def _get_tfonnx_emd_params(self):
        # Raises error if framework specified is TF-ONNX but is not supported by the model
        raise NotImplementedError(
            "TF-ONNX framework is currently not supported by this model."
        )

    def _save_as_tfonnx(self, saved_path, batch_size):
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                import onnx
                from onnx_tf.backend import prepare
        except:
            raise Exception(
                "Could not find the required deep learning dependencies. Ensure you have installed the required dependent libraries. See https://developers.arcgis.com/python/guide/deep-learning/."
            )

        batch_size = int(math.sqrt(int(batch_size))) ** 2
        dummy_input = torch.randn(
            batch_size,
            3,
            self._data.chip_size,
            self._data.chip_size,
            device=self._device,
            requires_grad=True,
        )
        torch.onnx.export(
            self.learn.model, dummy_input, saved_path.with_suffix(".onnx")
        )

    def save(
        self,
        name_or_path,
        framework="PyTorch",
        publish=False,
        gis=None,
        compute_metrics=True,
        save_optimizer=False,
        save_inference_file=True,
        **kwargs,
    ):
        """
        Saves the model weights, creates an Esri Model Definition and Deep
        Learning Package zip for deployment to Image Server or ArcGIS Pro.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Name of the model to save. It
                                stores it at the pre-defined location. If path
                                is passed then it stores at the specified path
                                with model name as directory name and creates
                                all the intermediate directories.
        ---------------------   -------------------------------------------
        framework               Optional string. Exports the model in the
                                specified framework format ('PyTorch', 'tflite'
                                'torchscript', and 'TF-ONXX' (deprecated)).
                                Only models saved with the default framework
                                (PyTorch) can be loaded using `from_model`.
                                ``tflite`` framework (experimental support) is
                                supported by :class:`~arcgis.learn.SingleShotDetector`
                                - tensorflow backend only,
                                :class:`~arcgis.learn.FeatureClassifier`(not supported
                                for transformer backbones) and
                                :class:`~arcgis.learn.RetinaNet` - tensorflow
                                backend only.``torchscript`` format is supported by
                                :class:`~arcgis.learn.SiamMask`,
                                :class:`~arcgis.learn.MaskRCNN`,
                                :class:`~arcgis.learn.SingleShotDetector`,
                                :class:`~arcgis.learn.YOLOv3` and
                                :class:`~arcgis.learn.RetinaNet`.
                                For usage of SiamMask model in ArcGIS Pro >= 2.8,
                                load the ``PyTorch`` framework saved model
                                and export it with ``torchscript`` framework
                                using ArcGIS API for Python >= v1.8.5.
                                For usage of SiamMask model in ArcGIS Pro 2.9,
                                set framework to ``torchscript`` and use the
                                model files additionally generated inside
                                'torch_scripts' folder.
                                If framework is ``TF-ONNX`` (Only supported for
                                :class:`~arcgis.learn.SingleShotDetector`),
                                ``batch_size`` can be passed as an optional
                                keyword argument.
        ---------------------   -------------------------------------------
        publish                 Optional boolean. Publishes the DLPK as an item.
        ---------------------   -------------------------------------------
        gis                     Optional :class:`~arcgis.gis.GIS`  Object.
                                Used for publishing the item. If not specified
                                then active gis user is taken.
        ---------------------   -------------------------------------------
        compute_metrics         Optional boolean. Used for computing model
                                metrics.
        ---------------------   -------------------------------------------
        save_optimizer          Optional boolean. Used for saving the model-optimizer
                                state along with the model. Default is set to False
        ---------------------   -------------------------------------------
        save_inference_file     Optional boolean. Used for saving the inference file
                                along with the model.
                                If False, the model will not work with ArcGIS Pro 2.6
                                or earlier. Default is set to True.
        ---------------------   -------------------------------------------
        kwargs                  Optional Parameters:
                                Boolean `overwrite` if True, it will overwrite
                                the item on ArcGIS Online/Enterprise, default False.
        =====================   ===========================================
        """
        if int(os.environ.get("RANK", 0)):
            return
        return self._save(
            name_or_path,
            framework=framework,
            publish=publish,
            gis=gis,
            compute_metrics=compute_metrics,
            save_optimizer=save_optimizer,
            save_inference_file=save_inference_file,
            **kwargs,
        )

    def load(self, name_or_path, **kwargs):
        """
        Loads a compatible saved model for inferencing or fine tuning from the disk.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Name or Path to
                                Deep Learning Package (DLPK) or
                                Esri Model Definition(EMD) file.
        =====================   ===========================================
        """
        temp = self.learn.path
        if "\\" in name_or_path or "/" in name_or_path:
            path = Path(name_or_path)
            # to make fastai from both path and with name
            if path.is_file():
                name = path.stem
                self.learn.path = path.parent
            else:
                name = path.parts[-1]
                self.learn.path = path
            self.learn.model_dir = ""
        else:
            # fixing fastai bug
            # self.learn.path = self.learn.path.parent
            self.learn.model_dir = Path(self.learn.model_dir) / name_or_path
            name = name_or_path

        try:
            device = getattr(self, "_map_location", None)
            self.learn.load(name, purge=False, device=device)
        except Exception as e:
            raise e
        finally:
            # undoing changes to self.learn.path
            self.learn.path = temp
            self.learn.model_dir = "models"
